/* $Id: attachableobject.cpp,v 1.10 2005/04/15 11:57:37 uwe Exp $ */

#include "attachableobject.hpp"
#include "serialize.hpp"
#include "constants.hpp"

/******************************************************************************/

void AttachableObject::printAttachInfo( const bool withOwnID ) const {
	INFO( "AttachableObject::printAttachInfo: info for %s (%p)\n",
	      withOwnID ? ObjectIDString[getID()] : "[n/a]", this );

	INFO( "\tnAttachedObjects: %i\n", getNAttachedObjects() );

	for( int i = 0; i < getNAttachedObjects(); i++ ) {
		INFO( "\tobject %2i is a %s (%p)\n", i,
		      ObjectIDString[getAttachedObject( i )->getID()],
		      getAttachedObject( i ) );
	}
}

/******************************************************************************/

void AttachableObject::serialize( Uint8*& bufferPointer ) const {
	// expands to a check of the buffer movement
	START_OBJECT_SERIALIZED_SIZE_CHECK( bufferPointer );

	Object::serialize( bufferPointer );	

	Serialize<const AttachableObject*>::serialize( this, bufferPointer );
	const Sint32 nAttachedObjects = m_attachedTo.size();
	Serialize<Sint32>::serialize( nAttachedObjects, bufferPointer );
	for( int i = 0; i < nAttachedObjects; i++ ) {
		Serialize<AttachableObject*>::serialize( m_attachedTo[i], bufferPointer );
	}

	// expands to tag serialization
	SERIALIZE_OBJECT_TAG( bufferPointer );
	// expands to a check of the buffer movement
	END_OBJECT_SERIALIZED_SIZE_CHECK( bufferPointer, AttachableObject );
}

/******************************************************************************/

void AttachableObject::deserialize( Uint8*& bufferPointer ) {
	Object::deserialize( bufferPointer );

	AttachableObject* oldPtr;

	Serialize<AttachableObject*>::deserialize( bufferPointer, oldPtr );
	Sint32 nAttachedObjects;
	Serialize<Sint32>::deserialize( bufferPointer, nAttachedObjects );
	for( int i = 0; i < nAttachedObjects; i++ ) {
		AttachableObject* o;
		Serialize<AttachableObject*>::deserialize( bufferPointer, o );
		m_attachedTo.push_back( o );
	}

	// expands to tag deserialization
	DESERIALIZE_OBJECT_TAG( bufferPointer );

	if( m_attachedTo.size() > 0 ) // only register if attached to other objects
		AttachableRegistrar::getInstance()->registerMe( oldPtr, this );
}

/******************************************************************************/

Uint32 AttachableObject::getSerializeBufferSize() const {
	return   Object::getSerializeBufferSize()
	       + Serialize<Sint32>::sizeOf()
	       + (1 + m_attachedTo.size()) * Serialize<AttachableObject*>::sizeOf()
	       // adds additional size for debugging (see serialize.hpp),
	       // but only, if the tags are used
	       PLUS_TAG_SIZE( 1 );
}

/******************************************************************************/
/******************************************************************************/
/******************************************************************************/

AttachableRegistrar::AttachableRegistrar() {}

/******************************************************************************/

AttachableRegistrar* AttachableRegistrar::m_instance = 0x0;
	
AttachableRegistrar* AttachableRegistrar::getInstance() {
	if( ! m_instance ) {
		m_instance = new AttachableRegistrar;
	}
	return m_instance;
}

/******************************************************************************/

void AttachableRegistrar::deleteInstance() {
	delete m_instance;
	m_instance = 0x0;
}

/******************************************************************************/

bool AttachableRegistrar::registerMe( const AttachableObject* oldPtr,
                                      AttachableObject* currentPtr ) {
	LOG( 3 ) INFO( "AttachableRegistrar::registerMe: registering AO\n" );
	return m_mapIDToPtr.insert( std::make_pair( oldPtr, currentPtr ) ).second;
}

/******************************************************************************/

bool AttachableRegistrar::startPointerConversion() {
	LOG( 2 ) INFO( "AttachableRegistrar::startPointerConversion: converting %i "
	                "AOs\n", m_mapIDToPtr.size() );

	std::map<const AttachableObject*, AttachableObject*>::iterator it;
	for( it = m_mapIDToPtr.begin(); it != m_mapIDToPtr.end(); it++ ) {
		LOG( 2 ) INFO( "AttachableRegistrar::startPointerConversion: converting\n" );

		std::vector<AttachableObject*>& ao = it->second->getAttachedObjects();
		for( unsigned int o = 0; o < ao.size(); o++ ) {
			std::map<const AttachableObject*, AttachableObject*>::iterator foundPtr =
			  m_mapIDToPtr.find( ao[o] );
			ASSERT( foundPtr != m_mapIDToPtr.end(),
			        "AttachableRegistrar::startPointerConversion: could not find object\n" );
			ao[o] = foundPtr->second;
		}

		LOG( 2 ) INFO( "AttachableRegistrar::startPointerConversion: converting done\n" );
	}

	return true;
}

/******************************************************************************/

void AttachableRegistrar::reset() {
	m_mapIDToPtr.clear();
	LOG( 2 ) INFO( "AttachableRegistrar::reset: done\n" );
}
