/* $Id: ball.cpp,v 1.14 2005/04/08 13:50:42 pohlt Exp $ */

#include "ball.hpp"
#include "world.hpp"
#include "particles.hpp"
#include "goal.hpp"
#include "rope.hpp"
#include "random.hpp"
#include "color.hpp"

/**********************************************************/

Mix_Chunk* Ball::m_explodeSample, * Ball::m_bounceSample,
         * Ball::m_damageSample;

/**********************************************************/

Ball::Ball() {
	m_health = 1;
	m_mass = 50.0;
	m_collRectX = -3; m_collRectY = -3;
	m_collRectWidth = 7; m_collRectHeight = 7;
	m_explodeSample = m_bounceSample = m_damageSample = NULL;
}

/**********************************************************/

Ball::~Ball() {
}

/**********************************************************/

bool Ball::initialize( const Vector& pos, const Vector& vel,
                       const Sint32 initialScoringDelay ) {
	int dummy;

	if( ! m_worldPointer->getMap()->testFilledCollRect( ROUND( pos.x ) + m_collRectX,
	                                                    ROUND( pos.y ) + m_collRectY,
	                        m_collRectWidth, m_collRectHeight, dummy, dummy ) ) {
		m_worldPointer->deleteObject( this );
		// DO NOT TOUCH ANY BALL DATA MEMBERS AFTER THIS POINT
		return false;
	}

	m_positionX = pos.x;
	m_positionY = pos.y;

	m_speedX = vel.x;
	m_speedY = vel.y;

	m_initialScoringDelay = m_scoringDelay = initialScoringDelay;

	// set ball stencil
	setCollRect();

	return true;
}

/**********************************************************/

void Ball::update() {
	if( m_damageSampleTimeout > 0 ) m_damageSampleTimeout--;

	Map* const map = m_worldPointer->getMap();

	const real DT = 1.0;

	int cx, cy, pcx, pcy;

	// remove ball stencil
	removeCollRect();

	if( isGrounded() && ( ! testGrounded() || getSpeed().abs2() > 0.01) ) {
		setState( GROUNDED, false );
	}

	if( isGrounded() ) {
		m_speedX = m_speedY = 0;
	}
	else {
		// move ball
		m_speedY += GRAVITATION*DT;

		real newPosX = m_positionX+m_speedX*DT,
		      newPosY = m_positionY+m_speedY*DT;

		// is there something to collide?
		if( map->testLineCollRect( ROUND( m_positionX ) + m_collRectX,
		                           ROUND( m_positionY ) + m_collRectY,
		                           ROUND( newPosX     ) + m_collRectX,
		                           ROUND( newPosY     ) + m_collRectY,
		                           m_collRectWidth, m_collRectHeight,
		                           pcx, pcy, cx, cy ) ) {
			// no, then just keep on flying
			m_positionX = newPosX;
			m_positionY = newPosY;
		}
		else {
			// play bounce sound
			m_positionX = pcx - m_collRectX;
			m_positionY = pcy - m_collRectY;
			if( ! m_firstUpdate ) {
				Audio::getInstance()->playSound( m_bounceSample, ROUND( m_positionX ),
																												 ROUND( m_positionY ) );
				const real speed2 = m_speedX*m_speedX + m_speedY*m_speedY;

				if( speed2 < 5.0 && testGrounded() ) {
					setState( GROUNDED, true );
					m_speedX = m_speedY = 0;
				}
				else {
					const real speed = SQRT_REAL( speed2 ),
					            outAngle = map->calculateBounceAngle( m_speedX, m_speedY, cx, cy, Flags::OBSTACLE_EARTH | Flags::OBSTACLE_OBJECT );
					m_speedX = speed * (speed > 5.0 ? 0.5 : 0.9) * COS_REAL( outAngle );
					m_speedY = speed * (speed > 5.0 ? 0.5 : 0.9) * SIN_REAL( outAngle );
				}
			}
		}
	}

	// check for goals
	bool insideGoal = false;
	int ballOwner = -1;

	for( int a = 0; a < getNAttachedObjects(); a++ ) {
		if( ! insideGoal && getAttachedObject( a )->getID() == GOAL_50 &&
		    dynamic_cast<Goal*>( getAttachedObject( a ))->testInside( getPosition() )) {
			insideGoal = true;
		}
		if( getAttachedObject( a )->getID() == ROPE ) {
			const int ropeOwner = dynamic_cast<Rope*>( getAttachedObject( a ))->getOwner();
			if( ballOwner == -1 ) ballOwner = ropeOwner;
			else if( ballOwner != ropeOwner ) {
				ballOwner = -2;
				break;
			}
		}
	}

	if( insideGoal && ballOwner >= 0 ) {
		// generate 'rainbow clock' particles
		Vector particleSpeed;
		for( int p = 0; p < 10; p++ ) {
			particleSpeed.setAngle( 2.0 * M_PI * m_scoringDelay / m_initialScoringDelay, 5.0 );
			particleSpeed += globalRnd.getVector( 2.0 * globalRnd.getNormedReal() );
			m_worldPointer->getParticles()->addParticle( Particles::ParticleData(
			    getPosition(), particleSpeed,
			    Particles::BOUNCE_EARTH | Particles::STAIN | Particles::REDUNDANT, 0,
			    0, 75, 1, Color::HSVToUint32( static_cast<real>( m_scoringDelay )/ m_initialScoringDelay, 1.0, 1.0 )));
		}

		// has the ball been inside the goal long enough?
		if( --m_scoringDelay <= 0 ) {
			for( int p = 0; p < 1000; p++ ) {
				particleSpeed.setAngle( p / 500.0 * M_PI, 5.0 );
				particleSpeed += globalRnd.getVector( 2.0 * globalRnd.getNormedReal() );
				m_worldPointer->getParticles()->addParticle( Particles::ParticleData(
				    getPosition(), particleSpeed,
				    Particles::BOUNCE_EARTH | Particles::STAIN | Particles::REDUNDANT, 0,
				    0, 75, 1, Color::HSVToUint32( p / 1000.0, 1.0, 1.0 )));
			}

			LOG( 1 ) INFO( "Ball::update: player/team %i just scored a goal A\n", ballOwner );
			Scorekeeper::getInstance()->playerScoredGoal( ballOwner );
			LOG( 1 ) INFO( "Ball::update: player/team %i just scored a goal B\n", ballOwner );
			m_worldPointer->deleteObject( this );
			LOG( 1 ) INFO( "Ball::update: player/team %i just scored a goal C\n", ballOwner );
			// DO NOT TOUCH BALL AFTER THIS POINT!!!
			return;
		}
	}
	else m_scoringDelay = m_initialScoringDelay;

	// set ball stencil
	setCollRect();

	m_firstUpdate = false;

	DBG( 5 ) map->setColorAt( ROUND( m_positionX ),
	                          ROUND( m_positionY ), 0xffffff );
}

/**********************************************************/

int Ball::applyDamage( const Particles::ParticleData& p ) {
	const int damage = CollidableObject::applyDamage( p );

	if( damage > 50 && m_damageSampleTimeout == 0 ) {
		Audio::getInstance()->playSound( m_damageSample, ROUND( m_positionX ),
		                                                 ROUND( m_positionY ) );
		m_damageSampleTimeout = 10;
	}

	m_health = 1; // indestructible object

	return damage;
}

/**********************************************************/

void Ball::serialize( Uint8*& bufferPointer ) const {
	// expands to a check of the buffer movement
	START_OBJECT_SERIALIZED_SIZE_CHECK( bufferPointer );
	
	CollidableObject::serialize( bufferPointer );	

	Serialize<Sint32>::serialize( m_initialScoringDelay, bufferPointer );
	Serialize<Sint32>::serialize( m_scoringDelay       , bufferPointer );

	// expands to tag serialization
	SERIALIZE_OBJECT_TAG( bufferPointer );
	// expands to a check of the buffer movement
	END_OBJECT_SERIALIZED_SIZE_CHECK( bufferPointer, Ball );
}

/**********************************************************/

void Ball::deserialize( Uint8*& bufferPointer ) {
	CollidableObject::deserialize( bufferPointer );

	Serialize<Sint32>::deserialize( bufferPointer, m_initialScoringDelay );
	Serialize<Sint32>::deserialize( bufferPointer, m_scoringDelay        );

	// expands to tag deserialization
	INFO( "testing ball object tag\n" );
	DESERIALIZE_OBJECT_TAG( bufferPointer );
}

/**********************************************************/

/* most likely redundant */
Uint32 Ball::getSerializeBufferSize() const {
	return CollidableObject::getSerializeBufferSize()
	       + sizeof( m_initialScoringDelay )
	       + sizeof( m_scoringDelay )
	       PLUS_TAG_SIZE( 1 );
}

