/* $Id: clientcomm.cpp,v 1.40 2005/04/19 15:53:25 chfreund Exp $ */

#include "clientcomm.hpp"

#include <cstdlib>
#include <new>

#include "serialize.hpp"
#include "packet.hpp"
#include "world.hpp"
#include "client.hpp"

ClientComm::ClientComm( Client* client )
	: m_tcpSocket( NULL ), m_tcpSocketSet( NULL ), m_udpSocket( NULL ),
	  m_localPort( 0 ), m_client( client ), m_isConnected( false ) {
	LOG( 2 ) INFO( "ClientComm::ClientComm: Creating TCP socket set\n" );
	m_tcpSocketSet = SDLNet_AllocSocketSet( 1 );
	ASSERT( m_tcpSocketSet,
	        "ClientComm::ClientComm: Could not create TCP socket set (%s)\n",
	        SDLNet_GetError() );

	LOG( 2 ) INFO( "ClientComm::ClientComm: Creating UDP socket\n" );
	m_udpSocket = SDLNet_UDP_Open( 0 );
	ASSERT( m_udpSocket,
	        "ClientComm::ClientComm: Could not create UDP socket (%s)\n",
	        SDLNet_GetError() );
}

ClientComm::~ClientComm() {
	// try to close all remaining connections
	disconnect();

	// free resources
	if ( m_tcpSocketSet ) {
		SDLNet_FreeSocketSet( m_tcpSocketSet );
		m_tcpSocketSet = NULL;
	}
	if ( m_udpSocket ) {
		SDLNet_UDP_Close( m_udpSocket );
		m_udpSocket = NULL;
	}
}

// This method tries to find servers on the local subnet. The detected servers
// will be collected in the returned vector of server entries.

std::set<ServerEntry> ClientComm::lookForServer() {
	std::set<ServerEntry> servers;
	ServerEntry entry;
	LOG( 2 ) INFO( "ClientComm::lookForServer: Sending bait packets to subnet\n" );

	// one broadcast to all reachable addresses
	SDLNet_ResolveHost( &m_baitPacket->address, "255.255.255.255", m_SERVER_PORT );
	m_baitPacket->len = 4;
	LOG( 5 ) INFO( "ClientComm::lookForServer: Sending bait packet to 255.255.255.255:%i\n",
				   m_baitPacket->address.port );
	CHECK( SDLNet_UDP_Send( m_udpSocket, -1, m_baitPacket ) == 1,
		   "ClientComm::lookForServer: Could not send bait packet (%s)\n",
		   SDLNet_GetError() );
	
	// explicit check the local host
	SDLNet_ResolveHost( &m_baitPacket->address, "127.0.0.1", m_SERVER_PORT );
	m_baitPacket->len = 4;
	LOG( 5 ) INFO( "ClientComm::lookForServer: Sending bait packet to 127.0.0.1:%i\n",
	               m_baitPacket->address.port );
	CHECK( SDLNet_UDP_Send( m_udpSocket, -1, m_baitPacket ) == 1,
	      "ClientComm::lookForServer: Could not send bait packet (%s)\n",
	      SDLNet_GetError() );
	
	// let's wait a little for the incoming answers
	SDL_Delay( 500 );

	// check the bait socket for echo packets
	LOG( 2 ) INFO( "ClientComm::lookForServer: Listening for server responses\n" );
	while ( SDLNet_UDP_Recv( m_udpSocket, m_baitPacket ) == 1 ) {
		LOG( 2 ) INFO( "ClientComm::lookForServer: Received bait packet (len: %i) from %i.%i.%i.%i:%i\n",
		               m_baitPacket->len,
		               DECODE_IP( m_baitPacket->address.host ),
		               m_baitPacket->address.port );
		if( m_baitPacket->len == 4 &&
		    m_baitPacket->data[0] == 0xb8 && m_baitPacket->data[1] == 0xb8 ) {
			LOG( 2 ) INFO( "ClientComm::lookForServer: Valid server response\n" );
			// the server address is given in the packet info
			// m_serverAddress = m_baitPacket->address;
			// the echo packet contains the number of the local port where we
			// have sent the bait packet from (is there a better way?)
			Uint8* dataPointer = reinterpret_cast<Uint8*>( m_baitPacket->data+2 );
			Serialize<Uint16>::deserialize( dataPointer, m_localPort );
			LOG( 3 ) INFO("ClientComm::lookForServer: local port is %d\n", m_localPort );
			// create a server entry for the detected server and add it to list
			entry.address = m_baitPacket->address;
			const char* serverName = SDLNet_ResolveIP( &entry.address );
			if ( serverName ) {
				strncpy( entry.name, serverName, 256 );
			}
			
			// Try to connect to the server to get info about the running game
			IPaddress connectAddress = entry.address;
			Uint8* portPointer = reinterpret_cast<Uint8*>( &connectAddress.port );
			Serialize<Uint16>::serialize( m_SERVER_PORT, portPointer );
			TCPsocket socket = SDLNet_TCP_Open( &connectAddress );
			if ( !socket ) {
				continue;
			}
			
			// Send a INFO packet
			LOG( 2 ) INFO( "ClientComm::lookForServer: send INFO packet to server\n" );
			TCPInfoPacket infoPacket;
			if ( tcpSendPacket( socket, &infoPacket ) == -1 ) {
				continue;
			}
			
			// receive INFO packet
			LOG( 2 ) INFO( "ClientComm::lookForServer: receiving INFO packet\n" );
			TCPPacket* packet = tcpRecvPacket( socket );
			if ( packet ) {
				ASSERT( packet->request == TCPPacket::INFO,
						"ClientComm::lookForServer: protocol error: expected INFO packet\n" );
				TCPInfoPacket* infoPacket2 =
					static_cast<TCPInfoPacket*>( packet );
				// extract information from the info packet
				entry.theme = infoPacket2->theme;
				entry.numberPlayers = infoPacket2->numberPlayers;
				delete packet;
			}
			else {
				continue;
			}
			
			servers.insert( entry );
		}
		else {
			LOG( 2 ) INFO( "ClientComm::lookForServer: Invalid server response\n" );
		}
	}
	
	return servers;
}

// This method sets the game server.

bool ClientComm::setServer( ServerEntry& entry ) {
	// If no explicit port number is given use the default port number
	Uint16 port = 0;
	if ( entry.address.port == 0 ) {
		port = m_SERVER_PORT;
	}
	else {
		Uint8* buffer = reinterpret_cast<Uint8*>( &entry.address.port );
		Serialize<Uint16>::deserialize( buffer, port );
	}

	// If no explicit address is given we have to resolve it via the name
	if ( entry.address.host == 0 ) {
		// Check if a server name is provided
		if ( strlen( entry.name ) == 0 ) {
			LOG( 2 ) INFO( "ClientComm::setServer: no address and no name provided\n" );
			return false;
		}
		
		// Resolve the address from the given name
		LOG( 2 ) INFO( "ClientComm::setServer: resolve server address\n" );
		if ( SDLNet_ResolveHost( &entry.address, entry.name, port ) == -1 ) {
			LOG( 2 ) INFO( "ClientComm::setServer: could not resolve server address (%s)\n",
						   SDLNet_GetError() );
			return false;
		}
		LOG( 3 ) INFO( "ClientComm::setServer: server address is %i.%i.%i.%i\n", 
					   DECODE_IP( entry.address.host ) );
	}
	else {
		// If no name is provided try to resolve it
		const char* serverName = SDLNet_ResolveIP( &entry.address );
		if ( serverName ) {
			strncpy( entry.name, serverName, 256 );
		}
	}
	
	// Use the given server entry
	m_server = entry;
	
	// if our local port is still unknown we have to find it out using
	// a bait packet
	if ( m_localPort == 0 ) {
		m_baitPacket->address = m_server.address;
		m_baitPacket->len = 4;
		LOG( 5 ) INFO( "ClientComm::setServer: Sending bait packet to %i.%i.%i.%i:%i\n",
					   DECODE_IP( m_baitPacket->address.host ),
					   m_baitPacket->address.port );
		CHECK( SDLNet_UDP_Send( m_udpSocket, -1, m_baitPacket ) == 1,
			   "ClientComm::setServer: Could not send bait packet (%s)\n",
			   SDLNet_GetError() );
	
		SDL_Delay( 500 );
	
		// Hopefully the time was long enough for the echo packet to arrive
		LOG( 2 ) INFO( "ClientComm::setServer: Listening for server response\n" );
		if ( SDLNet_UDP_Recv( m_udpSocket, m_baitPacket ) == 1 ) {
			LOG( 2 ) INFO( "ClientComm::setServer: Received bait packet (len: %i) from %i.%i.%i.%i:%i\n",
						   m_baitPacket->len,
						   DECODE_IP( m_baitPacket->address.host),
						   m_baitPacket->address.port );
			if( m_baitPacket->len == 4 &&
				 m_baitPacket->data[0] == 0xb8 && m_baitPacket->data[1] == 0xb8 ) {
				LOG( 2 ) INFO( "ClientComm::setServer: Valid server response\n" );
				m_server.address = m_baitPacket->address;
				Uint8* dataPointer = reinterpret_cast<Uint8*>( m_baitPacket->data+2 );
				Serialize<Uint16>::deserialize( dataPointer, m_localPort );
				LOG( 3 ) INFO("ClientComm::setServer: local port is %d\n", m_localPort );
				return true;
			}
			else {
				LOG( 2 ) INFO( "ClientComm::setServer: Invalid server response\n" );
			}
		}
		
		LOG( 2 ) INFO( "ClientComm::setServer: No server response\n" );
		return false;
	}
	
	return true;
}

bool ClientComm::connect( Player* player, World*& world ) {
	// Open a socket to the server
	LOG( 2 ) INFO( "ClientComm::connect: open socket to server\n" );
	IPaddress connectAddress = m_server.address;
	Uint8* portPointer = reinterpret_cast<Uint8*>( &connectAddress.port );
	Serialize<Uint16>::serialize( m_SERVER_PORT, portPointer );
	m_tcpSocket = SDLNet_TCP_Open( &connectAddress );
	ASSERT ( m_tcpSocket, "ClientComm::connect: could not open socket (%s)\n",
	         SDLNet_GetError() );
	SDLNet_TCP_AddSocket( m_tcpSocketSet, m_tcpSocket );
			 
	// Send a CONNECT packet
	LOG( 2 ) INFO( "ClientComm::connect: send CONNECT packet to server\n" );
	TCPConnectPacket connectPacket;
	connectPacket.port = m_localPort;
	if ( tcpSendPacket( m_tcpSocket, &connectPacket ) == -1 ) {
		m_client->setStatus( SHUTDOWN );
		return false;
	}
	
	// send player
	LOG( 2 ) INFO( "ClientComm::connect: sending player to server\n" );
	TCPPlayerPacket playerPacket;
	playerPacket.player = player;
	if ( tcpSendPacket( m_tcpSocket, &playerPacket ) == -1 ) {
		m_client->setStatus( SHUTDOWN );
		return false;
	}
	
	// receive the world
	LOG( 2 ) INFO( "ClientComm::connect: receiving world\n" );
	TCPPacket* packet = tcpRecvPacket( m_tcpSocket );
	if ( packet ) {
		ASSERT( packet->request == TCPPacket::WORLD,
				"ClientComm::connect: protocol error: expected WORLD packet\n" );
		TCPWorldPacket* worldPacket =
			static_cast<TCPWorldPacket*>( packet );
		world = worldPacket->world;
		// extract other relevant information from the world packet
		player->setPlayerID( worldPacket->playerID );
		LOG( 1 ) INFO( "ClientComm:: got player id %i\n", player->getPlayerID() );
		m_client->setFrame( worldPacket->frame );
		m_lastReceivedFrame = worldPacket->frame - 1;
		delete packet;
	}
	else {
		m_client->setStatus( SHUTDOWN );
		return false;
	}
	
	m_isConnected = true;
	return true;
}

void ClientComm::disconnect() {
	LOG( 1 ) INFO( "ClientComm::disconnect: starting destruction\n" );

	if( m_tcpSocketSet && m_tcpSocket ) {
		SDLNet_TCP_DelSocket( m_tcpSocketSet, m_tcpSocket );
	}
	if ( m_tcpSocket ) {
		SDLNet_TCP_Close( m_tcpSocket );
		m_tcpSocket = NULL;
	}

	m_isConnected = false;
	LOG( 1 ) INFO( "ClientComm::disconnect: done\n" );
}

// This method sends the given event to the server

void ClientComm::sendEvent( const Event& event ) {
	LOG( 5 ) INFO( "ClientComm::sendEvent: Sending event to server\n" );

	// TODO: check which of these assignments really has to be done every time
	m_eventPacket->address = m_server.address;
	m_eventPacket->len     = m_eventPacket->maxlen = EventData::getStaticSerializeBufferSize();

	// fill out event data fields
	m_eventData.event = event;
	m_eventData.lastReceivedFrame = m_lastReceivedFrame;

	// dump event
	m_eventData.dumpToFile( "/tmp/client.log", m_client->getLocalPlayer()->getPlayerID() );

	// serialize event data into packet
	Uint8* bufferPointer = m_eventPacket->data;
	m_eventData.serialize( bufferPointer );

	CHECK( SDLNet_UDP_Send ( m_udpSocket, -1, m_eventPacket ) == 1,
	       "ClientComm::SendEvent: Couldn't send  packet to server (%s)\n",
	       SDLNet_GetError() );
}

// Receives an echo packet from the server

bool ClientComm::receiveEcho( EchoData& echo ) {
	LOG( 5 ) INFO( "ClientComm::receiveEcho: Receiving from server\n" );

	while ( SDLNet_UDP_Recv( m_udpSocket, m_echoPacket ) == 1 ) {
		// discard answers to bait packets
		if ( m_echoPacket->len == 4 &&
			  m_echoPacket->data[0] == 0xb8 && m_echoPacket->data[1] == 0xb8 ) {
			DBG( 2 ) INFO( "ClientComm::receiveEcho: Discarding bait packet\n" );
			continue;
		}
		
		LOG( 5 ) INFO( "ClientComm::receiveEcho: Received echo packet (len: %i)\n",
		               m_echoPacket->len );

		// deserialize packet
		Uint8* bufferPointer = m_echoPacket->data;
		echo.deserialize( bufferPointer );
		
		return true;
	}
	
	return false;
}

EchoData* ClientComm::getEcho( Uint8 frame ) {
	// Check if the wanted frame is already in the buffer
	// Quick return in that case
	if ( m_echoBuffer.find( frame ) != m_echoBuffer.end() ) {
		LOG( 3 ) INFO( "ClientComm::getEcho(): found echo for frame %d in buffer\n", frame );
		EchoData* echo = m_echoBuffer[frame];
		m_echoBuffer.erase( frame );
		m_lastReceivedFrame = frame;
		return echo;
	}
	
	// Else receive incoming events and hope for the best
	EchoData echo;
	while ( receiveEcho( echo ) ) {
		// Check if this is the wanted frame
		if ( echo.frame == frame ) {
			LOG( 5 ) INFO( "ClientComm::getEcho(): received the right echo packet for frame %d\n", frame );
			m_lastReceivedFrame = frame;
			return NEW EchoData( echo );
		}
		
		// Check if this an old frame
		// TODO: check the condition for that
		if ( ( ( static_cast<Sint16>( echo.frame ) - static_cast<Sint16>( frame ) )
		       + N_ECHO_BUFFERS ) % N_ECHO_BUFFERS > N_ECHO_BUFFERS/2 ) {
			LOG( 3 ) INFO( "ClientComm::getEcho(): received an old echo packet for frame %d (expecting %d\n", echo.frame, frame );
			// throw away data
			continue;
		}
		
		// Check if we already received that frame
		if ( m_echoBuffer.find( echo.frame ) != m_echoBuffer.end() ) {
			LOG( 3 ) INFO( "ClientComm::getEcho(): received an already received echo packet for frame %d\n", echo.frame );
			continue;
		}
		
		// Add data to buffer
		LOG( 3 ) INFO( "ClientComm::getEcho(): storing echo for frame %d in buffer (expecting %d)\n", echo.frame, frame );
		m_echoBuffer[echo.frame] = NEW EchoData( echo );
	}
	
	return 0;
}

void ClientComm::handleTCPMessages() {
	// check the socket set with the TCP socket for activity
	if ( SDLNet_CheckSockets( m_tcpSocketSet, 0 ) > 0 ) {
		// receive packet
		TCPPacket* packet = tcpRecvPacket( m_tcpSocket );
		// a null packet indicates that the TCP connection has been closed,
		// probably because of a server shutdown -> we will also quit
		if ( !packet ) {
			m_client->setStatus( SHUTDOWN );
			return;
		}
		
		// distinguish different packet types
		switch ( packet->request ) {
			case TCPPacket::PLAYER: {
				TCPPlayerPacket* playerPacket =
					static_cast<TCPPlayerPacket*>( packet );
				m_client->addPlayer( playerPacket->player, playerPacket->frame );
				break;
			}
			case TCPPacket::CHAT: {
				TCPChatPacket* chatPacket =
					static_cast<TCPChatPacket*>( packet );
				// m_client->addChatMessage( chatPacket->message );
				m_client->showMessage( chatPacket->message, chatPacket->color );
				break;
			}
			case TCPPacket::QUIT: {
				TCPQuitPacket* quitPacket =
					static_cast<TCPQuitPacket*>( packet );
				m_client->removePlayer( quitPacket->player, quitPacket->frame );
				break;
			}
			default:
				INFO( "ClientComm::handleTCPMessages: protocol exception: unexpected packet\n" );
				break;
		}
		
		delete packet;
	}
}

void ClientComm::sendChatMessage( String& message, const SDL_Color& color ) {
	TCPChatPacket chatPacket;
	chatPacket.message = message;
	chatPacket.color = color;
	LOG( 2 ) INFO( "ClientComm::sendChatMessage: send chat message to server\n" );
	if ( tcpSendPacket( m_tcpSocket, &chatPacket ) == -1 ) {
		m_client->setStatus( SHUTDOWN );
	}
}
