/* $Id: comm.cpp,v 1.39 2005/04/19 11:41:13 chfreund Exp $ */

// #include <cstdlib>
#include <cstdio>
#include <unistd.h>
#include <new>
#include <errno.h>
#ifdef WIN32
#include <winsock.h>
#endif
#include "global.hpp"
#include "comm.hpp"

Comm::Comm() {
	// Initialize the SDLnet
	LOG( 1 ) INFO( "Comm::Comm: Initializing SDLNet\n" );
	ASSERT( SDLNet_Init ()	!=	-1,
	        "Comm::Comm: Couldn't initialize SDLNet (%s)\n",
	        SDLNet_GetError() );

	// This should determine the own hostname
	// (is gethostname platform independent?)
	LOG( 2 ) INFO( "Comm::Comm: Resolve local hostname\n" );
	char hostname[256];
	ASSERT( gethostname( hostname, 256 ) == 0,
	        "Comm::Comm: could not resolve local hostname\n" );
	LOG( 3 ) INFO( "Comm::Comm: local hostname is %s\n", hostname );

	// Now we want to know our own IP address
	LOG( 2 ) INFO( "Comm::Comm: Resolving local address\n" );
	if ( SDLNet_ResolveHost( &m_localAddress, hostname, 0 ) == -1 ) {
		LOG( 3 ) INFO( "Comm::Comm: could not resolve local address for %s\n", hostname );
	}
	else {
		LOG( 3 ) INFO( "Comm::Comm: local address is %i.%i.%i.%i\n", DECODE_IP( m_localAddress.host ) );
	}

	LOG( 2 ) INFO( "Comm::Comm: Allocating memory for bait packet\n" );
	m_baitPacket = SDLNet_AllocPacket( 4 );
	ASSERT( m_baitPacket, "Comm::Comm: Could not allocate memory (%s)\n",
	        SDLNet_GetError() );
	// Initialize first two bytes of the bait packet with magic data
	m_baitPacket->data[0] = 0xb8;
	m_baitPacket->data[1] = 0xb8;

	LOG( 2 ) INFO( "Comm::Comm: Allocating memory for event packet\n" );
	m_eventPacket = SDLNet_AllocPacket( EventData::getStaticSerializeBufferSize() );
	ASSERT( m_eventPacket, "Comm::Comm: Could not allocate memory (%s)\n",
	        SDLNet_GetError() );

	LOG( 2 ) INFO( "Comm::Comm: Allocation memory for echo packet\n" );
	m_echoPacket = SDLNet_AllocPacket( EchoData::getStaticSerializeBufferSize() );
	ASSERT( m_echoPacket, "Comm::Comm: Could not allocate memory (%s)\n",
	        SDLNet_GetError() );
}

Comm::~Comm () {
	LOG( 1 ) INFO( "Comm::~Comm: starting destructor\n" );

	if ( m_baitPacket ) {
		SDLNet_FreePacket( m_baitPacket );
	}
	if ( m_eventPacket ) {
		SDLNet_FreePacket( m_eventPacket );
	}
	if ( m_echoPacket ) {
		SDLNet_FreePacket( m_echoPacket );
	}

	LOG( 1 ) INFO( "Comm::~Comm: shutting down SDLNet\n" );
	SDLNet_Quit ();

	LOG( 1 ) INFO( "Comm::~Comm: done\n" );
}

int Comm::tcpSendPacket( TCPsocket socket, TCPPacket* packet ) {
	// calculate data size header
	packet->dataSize = packet->getDataSize();
	
	// create send buffer
	Uint32 packetSize = packet->getSerializeBufferSize();
	Uint8* packetBuffer = NEW Uint8[packetSize];

	// serialize packet
	Uint8* bufferPointer = packetBuffer;
	packet->serialize( bufferPointer );
	
	// send packet
	Uint32 dataSent = 0;
	while ( dataSent < packetSize ) {
		int bytesSent = 
		     SDLNet_TCP_Send( socket, static_cast<void*>( packetBuffer + dataSent ),
			                    packetSize - dataSent );
		if ( bytesSent < (int)packetSize - (int)dataSent && errno != EAGAIN ) {
			LOG( 2 ) INFO( "Comm::tcpSendPacket: error sending packet (%s)\n",
			               SDLNet_GetError() );
			return -1;
		}
		dataSent += bytesSent;
	}
	delete[] packetBuffer;
	
	// error check
	if ( dataSent < packetSize ) {
		LOG( 2 ) INFO( "Comm::tcpSendPacket: error sending packet (%s)\n",
		               SDLNet_GetError() );
		return -1;
	}
	
	return 0;
}

TCPPacket* Comm::tcpRecvPacket( TCPsocket socket ) {
	// create packet for storing header data
	TCPPacket headerPacket;
	
	// create receive buffer
	Uint32 headerSize = TCPPacket::getStaticSerializeBufferSize();
	Uint8* headerBuffer = NEW Uint8[headerSize];
	
	// receive header data
	Uint32 dataReceived = 0;
	while ( dataReceived < headerSize ) {
		int bytesReceived =
			SDLNet_TCP_Recv( socket, headerBuffer + dataReceived,
			                 headerSize - dataReceived );
		if ( bytesReceived <= 0 ) {
			LOG( 2 ) INFO( "Comm::tcpRecvPacket: error receiving packet header (%s)\n",
			               SDLNet_GetError() );
			delete[] headerBuffer;
			return 0;
		}
		dataReceived += bytesReceived;
	}
	
	// deserialize header data
	Uint8* bufferPointer = headerBuffer;
	headerPacket.deserialize( bufferPointer );
	delete[] headerBuffer;
	

	// distinguish packet types
	TCPPacket* packet = 0;
	switch ( headerPacket.request ) {
		case TCPPacket::UNKNOWN:
			// nothing to do here
			break;
		case TCPPacket::INFO:
			packet = NEW TCPInfoPacket( headerPacket );
			break;
		case TCPPacket::CONNECT:
			packet = NEW TCPConnectPacket( headerPacket );
			break;
		case TCPPacket::PLAYER:
			packet = NEW TCPPlayerPacket( headerPacket );
			break;
		case TCPPacket::WORLD:
			packet = NEW TCPWorldPacket( headerPacket );
			break;
		case TCPPacket::CHAT:
			packet = NEW TCPChatPacket( headerPacket );
			break;
		case TCPPacket::QUIT:
			packet = NEW TCPQuitPacket( headerPacket );
		default:
			break;
	}
	
	// receive content data
	Uint8* dataBuffer = NEW Uint8[headerPacket.dataSize];
	dataReceived = 0;
	while ( dataReceived < headerPacket.dataSize ) {
		int bytesReceived =
			SDLNet_TCP_Recv( socket, dataBuffer + dataReceived,
			                 headerPacket.dataSize - dataReceived );
		if ( bytesReceived <= 0 ) {
			LOG( 2 ) INFO( "Comm::tcpRecvPacket: error receiving packet data (%s)\n",
			               SDLNet_GetError() );
			delete[] dataBuffer;
			delete packet;
			return 0;
		}
		dataReceived += bytesReceived;
	}
	
	// deserialize content data
	bufferPointer = dataBuffer;
	packet->deserializeData( bufferPointer );
	delete[] dataBuffer;
	
	return packet;
}

void Comm::tcpSendObject( TCPsocket socket, Serializable* object ) {
	// send object size
	Uint32 objectSize = object->getSerializeBufferSize();
	Uint8 sizeBuffer[4];
	Uint8* bufferPointer = sizeBuffer;
	Serialize<Uint32>::serialize( objectSize, bufferPointer );
	SDLNet_TCP_Send( socket, static_cast<void*>( sizeBuffer ), 4 );
	
	// send object data
	Uint8* objectBuffer = NEW Uint8[objectSize];
	bufferPointer = objectBuffer;
	object->serialize( bufferPointer );
	SDLNet_TCP_Send( socket, static_cast<void*>( objectBuffer ),
	                 objectSize );
	delete[] objectBuffer;
}

void Comm::tcpRecvObject( TCPsocket socket, Serializable* object ) {
	// receive object size
	Uint8 sizeBuffer[4];
	Uint32 dataReceived = 0;
	while ( dataReceived < 4 ) {
		dataReceived +=
			SDLNet_TCP_Recv( socket, sizeBuffer + dataReceived,
			                 4 - dataReceived );
	}
	Uint32 objectSize;
	Uint8* bufferPointer = sizeBuffer;
	Serialize<Uint32>::deserialize( bufferPointer, objectSize );
	
	// receive object data
	Uint8* objectBuffer = NEW Uint8[objectSize];
	dataReceived = 0;
	while ( dataReceived < objectSize ) {
		dataReceived +=
			SDLNet_TCP_Recv( socket, objectBuffer + dataReceived,
			                 objectSize - dataReceived );
	}
	bufferPointer = objectBuffer;
	object->deserialize( bufferPointer );
	delete[] objectBuffer;
}
