/* $Id: hook.cpp,v 1.18 2005/04/08 13:16:03 pohlt Exp $ */

#include "hook.hpp"
#include "world.hpp"
#include "avatar.hpp"
#include "rope.hpp"

/**********************************************************/

Mix_Chunk* Hook::m_attachSample, * Hook::m_damageSample;

/**********************************************************/

Hook::Hook() {
	m_health = 50000;
	m_mass = 15.0;
	m_collRectX = -1; m_collRectY = -1;
	m_collRectWidth = 3; m_collRectHeight = 3;
	m_stickingInEarth = false;
}

/**********************************************************/

bool Hook::initialize( const Vector& pos, const Vector& vel ) {
	int dummy;

	if( ! m_worldPointer->getMap()->testFilledCollRect( ROUND( pos.x ) + m_collRectX,
	                                                    ROUND( pos.y ) + m_collRectY,
	                        m_collRectWidth, m_collRectHeight, dummy, dummy ) ) {
		m_worldPointer->deleteObject( this );
		// DO NOT TOUCH ANY HOOK DATA MEMBERS AFTER THIS POINT
		return false;
	}

	m_positionX = pos.x;
	m_positionY = pos.y;

	m_speedX = vel.x;
	m_speedY = vel.y;

	// set hook stencil
	setCollRect();

	return true;
}

/**********************************************************/

void Hook::update() {
	if( m_damageSampleTimeout > 0 ) m_damageSampleTimeout--;

	Map* const map = m_worldPointer->getMap();

	const real DT = 1.0;

	int cx, cy, pcx, pcy;

	// remove hook stencil
	removeCollRect();

	DBG( 5 ) INFO( "Hook::update: collRect removed\n" );

	// if nothing is attached or hook is destroyed
	if( getNAttachedObjects() == 0 || m_health <= 0 ) {
		// remove hook object
		m_worldPointer->deleteObject( this );
		// DO NOT TOUCH ANY HOOK DATA MEMBERS AFTER THIS POINT
		return;
	}

	if( m_stickingInEarth ) {
		if( map->testPassableSmoothing( m_stickingAtX,
		                                m_stickingAtY,
		                                Flags::OBSTACLE_EARTH ) ) {
			m_stickingInEarth = false;
		}
	}
	
	if( m_stickingInEarth ) {
		m_speedX = m_speedY = 0;
	}
	else {
		// move hook
		m_speedY += GRAVITATION*DT;

		real newPosX = m_positionX+m_speedX*DT,
		      newPosY = m_positionY+m_speedY*DT;

		// is there something to collide?
		if( map->testLineCollRect( ROUND( m_positionX ) + m_collRectX,
		                           ROUND( m_positionY ) + m_collRectY,
		                           ROUND( newPosX     ) + m_collRectX,
		                           ROUND( newPosY     ) + m_collRectY,
		                           m_collRectWidth, m_collRectHeight,
		                           pcx, pcy, cx, cy ) ) {
			// no, then just keep on flying
			m_positionX = newPosX;
			m_positionY = newPosY;
		}
		else {
			// yes
			m_positionX = pcx - m_collRectX;
			m_positionY = pcy - m_collRectY;
			//Audio::getInstance()->playSound( m_attachSample, ROUND( m_positionX ),
			//                                                 ROUND( m_positionY ) );

			const Flag hitFlag = map->getFlags()->getValueAt( cx, cy );

			if( hitFlag & Flags::OBSTACLE_EARTH ) {
				INFO( "Hook::update: hit earth\n" );
				m_stickingAtX = cx;
				m_stickingAtY = cy;
				m_speedX = m_speedY = 0;
				m_stickingInEarth = true;

				// set rope length to current length
				for( int i = 0; i < getNAttachedObjects(); i++ ) {
					Rope* rope = dynamic_cast<Rope*>( getAttachedObject( i ) );
					if( rope ) {
						rope->setLength( 0.85 * rope->getLength() );
					}
				}
			}

			if( hitFlag & Flags::OBSTACLE_OBJECT ) {
				INFO( "Hook::update: hit object\n" );

				CollidableObject* hitObject = m_worldPointer->getCollidableAt( cx, cy );
				if( hitObject ) {
					for( int i = 0; i < getNAttachedObjects(); i++ ) {
						const Rope* rope = dynamic_cast<Rope*>( getAttachedObject( i ));
						DBG( 1 ) CHECK( rope, "Hook::update: attached object is not a rope\n" );

						/* only attach if not yet attached */
						if( getAttachedObject( i )->getAttachedObjectIndex( hitObject ) == -1 ) {
							getAttachedObject( i )->attachTo( hitObject );
						}
						else {
							/* Risky to delete the rope because this would mess up the vector
							   containing the attached ropes. Better let the rope find out
							   that this hook has died and kill itself.
							delete getAttachedObject( i );
							*/
						}
					}
				}

				// remove hook object
				m_worldPointer->deleteObject( this );
				// DO NOT TOUCH ANY HOOK DATA MEMBERS AFTER THIS POINT
				return;
			}

			CHECK( hitFlag & Flags::OBSTACLE_ANY, "Hook::update: hit unknown type\n" );
		}
	}

	// set hook stencil
	setCollRect();

	DBG( 5 ) INFO( "Hook::update: collRect set\n" );
}

/**********************************************************/

int Hook::applyDamage( const Particles::ParticleData& p ) {
	const int damage = CollidableObject::applyDamage( p );

	if( damage > 50 && m_damageSampleTimeout == 0 ) {
		Audio::getInstance()->playSound( m_damageSample, ROUND( m_positionX ),
		                                                 ROUND( m_positionY ) );
		m_damageSampleTimeout = 10;
	}

	LOG( 5 ) INFO( "Hook::applyDamage: health: %i\n", m_health );

	return damage;
}

/**********************************************************/

void Hook::serialize( Uint8*& bufferPointer ) const {
	// expands to a check of the buffer movement
	START_OBJECT_SERIALIZED_SIZE_CHECK( bufferPointer );

	CollidableObject::serialize( bufferPointer );

	Serialize<bool>::serialize( m_stickingInEarth, bufferPointer );
	Serialize<Sint32>::serialize( m_stickingAtX, bufferPointer );
	Serialize<Sint32>::serialize( m_stickingAtY, bufferPointer );

	// expands to tag serialization
	SERIALIZE_OBJECT_TAG( bufferPointer );
	// expands to a check of the buffer movement
	END_OBJECT_SERIALIZED_SIZE_CHECK( bufferPointer, Hook );
}

/**********************************************************/

void Hook::deserialize( Uint8*& bufferPointer ) {
	CollidableObject::deserialize( bufferPointer );

	Serialize<bool>::deserialize( bufferPointer, m_stickingInEarth );
	Serialize<Sint32>::deserialize( bufferPointer, m_stickingAtX );
	Serialize<Sint32>::deserialize( bufferPointer, m_stickingAtY );

	// expands to tag deserialization
	DESERIALIZE_OBJECT_TAG( bufferPointer );
}

/**********************************************************/

Uint32 Hook::getSerializeBufferSize() const {
	return   CollidableObject::getSerializeBufferSize()
	       + Serialize<bool>::sizeOf( m_stickingInEarth )
	       + 2 * Serialize<Sint32>::sizeOf( m_stickingAtX )
	         // adds additional size for debugging (see serialize.hpp),
	         // but only, if the tags are used
	         PLUS_TAG_SIZE( 1 );
}

/**********************************************************/
