/* $Id: server.cpp,v 1.28 2005/03/28 20:03:41 pohlt Exp $ */

#include "server.hpp"

#include <new>
#include <SDL.h>

#include "wopsettings.hpp"
#include "world.hpp"
#include "servercomm.hpp"
#include "signalhandler.hpp"

#include "spritesequence.cpp"
#include "spriteset.cpp"
#include "wopsprites.cpp"

/**********************************************************/

// So far this function does not do anything sensible any more.
// It was introduced to allow the correct increment of all frame
// counters on the server, too. But the increment of the frame indices
// has now moved into the world, more precisely into the update routine
// of the objects. Nevertheless we do not delete this function, because
// we may need it once again.
void Server::updateGraphics()
{
	// draw the collidable objects on NULL screen
	for( unsigned int i = 0; i < m_world->getCollidables()->size(); i++ ) {
		m_sprites.draw( *(*m_world->getCollidables())[i],
		                 (*m_world->getCollidables())[i]->getIntPositionX(),
		                 (*m_world->getCollidables())[i]->getIntPositionY(),
		                 NULL );
	}
	// draw the noncollidable objects on NULL screen
	for( unsigned int i = 0; i < m_world->getNoncollidables()->size(); i++ ) {
		m_sprites.draw( *(*m_world->getNoncollidables())[i],
		                 (*m_world->getNoncollidables())[i]->getIntPositionX(),
		                 (*m_world->getNoncollidables())[i]->getIntPositionY(),
		                 NULL );
	}
}

/**********************************************************/

Server::Server() {
	// get WoP configuration
	WopSettings* settings = WopSettings::getInstance();

	// get map dimensions
	Uint32 sizeX = static_cast<Uint32>( settings->getWidth() );
	Uint32 sizeY = static_cast<Uint32>( settings->getHeight() );

	// create the world
	m_world = NEW World( sizeX, sizeY );
	ASSERT( m_world, "Server::Server: Could not create the world\n" );
	// connect the sprite class with the world using
	// the inteface base class of WopSprites
	m_world->setSpriteInterface( &m_sprites );
	
	// create the world from the theme in settings
	m_world->createFromTheme();
	
	// create the dummy sprite graphics
	ASSERT( m_sprites.load(), "Server::Server: could not create "
	        "graphic structure\n" );

	// setup game mode
	m_world->setupGameMode();

	m_replay = new GameReplay( "/tmp/wopreplay.dat", m_world, &globalRnd );
	ASSERT( m_replay, "Server::Server: Could not open replay file\n" );
	
	// create the mutex for client synchronisation
	m_syncMutex = SDL_CreateMutex();
	ASSERT( m_syncMutex, "Server::Server: Could not create mutex (%s)\n",
	        SDL_GetError() );
	// create the condition variable for client synchronisation
	m_syncCond = SDL_CreateCond();
	ASSERT( m_syncCond,
	        "Server::Server: Could not create condition variable (%s)\n",
	        SDL_GetError() );
	
	m_comm = NEW ServerComm( this, m_world );
	ASSERT( m_comm,
	        "Server::Server: Could not create server communicator (%s)\n",
	        SDL_GetError() );

	m_frame = 0;
	m_realFrame = 0;
}

/**********************************************************/

Server::~Server() {
	delete m_replay;
	delete m_comm;
	SDL_DestroyCond( m_syncCond );
	SDL_DestroyMutex( m_syncMutex );
	delete m_world;
}

/**********************************************************/

void Server::run() {
	SignalHandler* signalHandler = SignalHandler::getInstance();
	while( !signalHandler->hasQuit() ) {
		makePace();
		processEvents();
		if( signalHandler->hasSignal() ) {
			INFO( "Server::run: Signal received\n" );
			signalHandler->setSignal( false );
			if( signalHandler->hasPause() ) {
				INFO( "Server::run: Pausing. Press CTRL-Z to continue\n" );
				while( signalHandler->hasPause() ) {
					SDL_Delay( 1000 );
					LOG( 2 ) INFO( "Server::run: Pausing. Press CTRL-Z to continue\n" );
				}
			}
		}
	}
}

/**********************************************************/

void Server::processEvents() {
	// As long a we are in DEFAULT state we just collect the events from the
	// client, send echo messages back and update the world
	if ( m_comm->getStatus() == DEFAULT ) {
		m_comm->collectEvents();
		m_comm->broadcastEcho();
		
		const EchoData* echoData = m_comm->getEchoData();
		m_replay->addEcho( *echoData );
		
		if ( !( m_realFrame % 25  ) ) {
			DBG( 4 ) INFO( "Server::processEvents: starting frame %d\n", m_realFrame );
		}
		
		// Apply all events to the world
		m_world->applyEvents( echoData );
		// make time step in world
		m_world->doTimestep();
		// apply dummy drawing to update objects
		updateGraphics();
		
		// Check for TCP messages
		m_comm->handleTCPMessages();
		
		// Advance to next frame
		m_frame++;
		m_realFrame++;
	}
	// In SYNC state we stop receiving events and sending echoes
	else if ( m_comm->getStatus() == SYNC ) {
		// tell the client listening thread that we are now ready to perform the sync
		DBG( 2 ) INFO( "Server::processEvents: acquiring sync mutex\n" );
		SDL_LockMutex( m_syncMutex );
		DBG( 2 ) INFO( "Server::processEvents: signalling condition variable\n" );
		SDL_CondSignal( m_syncCond );
		DBG( 2 ) INFO( "Server::processEvents: releasing sync mutex\n" );
		SDL_UnlockMutex( m_syncMutex );
		
		// now we have to wait until the synchronisation is finished
		DBG( 2 ) INFO( "Server::processEvents: acquiring sync mutex\n" );
		SDL_LockMutex( m_syncMutex );
		DBG( 2 ) INFO( "Server::processEvents: waiting on condition variable\n" );
		SDL_CondWait( m_syncCond, m_syncMutex );
		DBG( 2 ) INFO( "Server::processEvents: releasing sync mutex\n" );
		SDL_UnlockMutex( m_syncMutex );
	}
}

int serverRunWrapper( void* data ) {
	Server* server = static_cast<Server*>( data );
	server->run();
	return 0;
}

/**********************************************************/
