/* $Id: servercomm.cpp,v 1.48 2005/05/09 20:04:52 chfreund Exp $ */

#include "servercomm.hpp"

#ifdef LOSSY_UDP_COMMUNICATION
#include <cstdlib>
#endif
#include <cstring>
#include <new>
#include <SDL.h>
#include <SDL_thread.h>
#include <SDL_net.h>

#include "serialize.hpp"
#include "packet.hpp"
#include "server.hpp"
#include "world.hpp"
#include "gamereplay.hpp"
#include "constants.hpp"
#include "wopsettings.hpp"

ServerComm::ServerComm( Server* server, World* world )
	: m_baitSocket( 0 ),
	  m_gameSocket( 0 ),
	  m_listenSocket( 0 ),
	  m_baitThread( 0 ),
	  m_listenThread( 0 ),
	  m_server( server ),
	  m_world( world ) {

	LOG( 1 ) INFO( "ServerComm::ServerComm: initializing\n" );
	
	// Resolve the address the server is supposed to listen on
	LOG( 2 ) INFO( "ServerComm::ServerComm: resolving listening address\n" );
	IPaddress listenAddress;
	SDLNet_ResolveHost( &listenAddress, 0, m_SERVER_PORT );	

	// Create the TCP socket
	LOG( 2 ) INFO( "ServerComm::ServerComm: creating listen socket\n" );
	m_listenSocket = SDLNet_TCP_Open( &listenAddress );
	ASSERT( m_listenSocket, "ServerComm::ServerComm: could not create listen socket (%s)\n",
	        SDLNet_GetError() );
	
	// Create the UDP socket for the in-game communication
	LOG( 2 ) INFO( "ServerComm::ServerComm: creating game socket\n" );
	m_gameSocket = SDLNet_UDP_Open( 0 );
	ASSERT( m_gameSocket, "ServerComm::ServerComm: could not create game socket (%s)\n",
	        SDLNet_GetError() );

	// Create the socket set which makes it easier to detect socket activity
	LOG( 2 ) INFO( "ServerComm::ServerComm: Allocating game set\n" );
	m_gameSet = SDLNet_AllocSocketSet( 1 );
	ASSERT( m_gameSet, "ServerComm::ServerComm: Could not allocate game set (%s)\n",
	        SDLNet_GetError() );
	
	LOG( 2 ) INFO( "ServerComm::ServerComm: Adding game socket to game set\n" );
	ASSERT( SDLNet_UDP_AddSocket( m_gameSet, m_gameSocket ) == 1,
	        "ServerComm::ServerComm: Could not add game socket to game set\n",
	        SDLNet_GetError() );

	// Create the UDP socket for the server discovery
	LOG( 2 ) INFO( "ServerComm::ServerComm: creating bait socket\n" );
	m_baitSocket = SDLNet_UDP_Open( m_SERVER_PORT );
	ASSERT( m_baitSocket, "ServerComm::ServerComm: could not create bait socket (%s)\n",
	        SDLNet_GetError() );
	
	LOG( 2 ) INFO( "ServerComm::ServerComm: Allocating bait set\n" );
	m_baitSet = SDLNet_AllocSocketSet( 1 );
	ASSERT( m_baitSet, "ServerComm::ServerComm: Could not allocate bait set (%s)\n",
	        SDLNet_GetError() );
	
	LOG( 2 ) INFO( "ServerComm::ServerComm: Adding bait socket to bait set\n" );
	ASSERT( SDLNet_UDP_AddSocket( m_baitSet, m_baitSocket ) == 1,
	        "ServerComm::ServerComm: Could not add bait socket to bait set\n",
	        SDLNet_GetError() );
	
	LOG( 2 ) INFO( "ServerComm::ServerComm: Allocating TCP set\n" );
	m_tcpSet = SDLNet_AllocSocketSet( MAX_NUMBER_OF_PLAYERS );
	ASSERT( m_tcpSet, "ServerComm::ServerComm: Could not allocate TCP set (%s)\n",
	        SDLNet_GetError() );
	
	m_status = DEFAULT;
	
	// Start listen thread
	LOG( 2 ) INFO( "ServerComm::ServerComm: Start listen thread\n" );
	m_listenThread = SDL_CreateThread( listenForClientsWrapper, this );
	ASSERT( m_listenThread, "ServerComm::ServerComm: Could not create listen thread (%s)\n",
	        SDL_GetError() );

	// Start bait thread
	LOG( 2 ) INFO( "ServerComm::ServerComm: Start bait thread\n" );
	m_baitThread = SDL_CreateThread( listenForBaitPacketsWrapper, this );
	ASSERT( m_baitThread, "ServerComm::ServerComm; Could not create bait thread (%s)\n",
	        SDL_GetError() );
}

ServerComm::~ServerComm() {
	LOG( 1 ) INFO( "ServerComm::~ServerComm: destroying\n" );

	// A thorough cleanup
	m_status = SHUTDOWN;
	
	// close connection to all clients
	for ( std::vector<Client*>::iterator it = m_client.begin();
	      it != m_client.end(); it++ ) {
		  Client* client = *it;
		  SDLNet_TCP_Close( client->socket );
	}
	
	if ( m_baitThread ) {
		SDL_WaitThread( m_baitThread, 0 );
		m_baitThread = 0;
	}
	if ( m_listenThread ) {
		SDL_WaitThread( m_listenThread, 0 );
		m_listenThread = 0;
	}
	if ( m_gameSet ) {
		SDLNet_FreeSocketSet( m_gameSet );
	}
	if ( m_baitSet ) {
		SDLNet_FreeSocketSet( m_baitSet );
	}
	if ( m_tcpSet ) {
		SDLNet_FreeSocketSet( m_tcpSet );
	}
	if ( m_gameSocket ) {
		SDLNet_UDP_Close( m_gameSocket );
	}
	if ( m_baitSocket ) {
		SDLNet_UDP_Close( m_baitSocket );
	}
	if ( m_listenSocket ) {
		SDLNet_TCP_Close( m_listenSocket );
	}
}

// In an infinite loop, this method checks the bait socket for incoming
// bait packets from clients which want to discover the server. For each
// incoming packet an echo packet is sent back to the client. The echo
// packet contains the port number of the client socket where the bait
// packets came from (this is one of the obviously very limited ways for the
// client to get to know the port number of the anonymously created game
// socket)

void ServerComm::listenForBaitPackets() {
	while ( m_status != SHUTDOWN ) {
		// Check the socket set for activity
		LOG( 5 ) INFO( "ServerComm::listenForBaitPackets: Listening for bait packets\n" );
		int numberActive = SDLNet_CheckSockets( m_baitSet, 0 );
		
		if ( numberActive > 0 ) {
			// Receive the incoming packet
			if ( SDLNet_UDP_Recv( m_baitSocket, m_baitPacket) == 1 ) {
				LOG( 2 ) INFO( "ServerComm::listenForBaitPackets: Received bait packet (len: %i) from %i.%i.%i.%i:%i\n",
				               m_baitPacket->len,
				               DECODE_IP( m_baitPacket->address.host), m_baitPacket->address.port);
				// Check if this is a valid packet and send an echo back to the client
				if ( m_baitPacket->len == 4 ) {
					LOG( 2 ) INFO( "ServerComm::listenForBaitPackets: Valid client request\n" );
					// Uint8* dataPointer = reinterpret_cast<Uint8*>( m_baitPacket->data );
					// Serialize<Uint16>::serialize( m_baitPacket->address.port, dataPointer );
					*(reinterpret_cast<Uint16*>(m_baitPacket->data+2)) = m_baitPacket->address.port;
					if ( CHECK( SDLNet_UDP_Send( m_gameSocket, -1, m_baitPacket ) == 1,
					            "ServerComm::listenForBaitPackets: Could not send bait packet (%s)\n",
					            SDLNet_GetError()) ) {
					}
				}
				else {
					LOG( 2 ) INFO( "ServerComm::listenForBaitPackets: Invalid client request (length = %d)\n",
					               m_baitPacket->len );
				}
			}
		}

		SDL_Delay( 500 );
	}
}

// Again in an infinite loop this method listens for incoming client
// connections. We choose to use a TCP connection because we have to send
// all the world data to the client.

void ServerComm::listenForClients() {
	// check for valid server socket
	if ( ! m_listenSocket ) {
		LOG( 2 ) INFO( "ServerComm::listenForClients: server socket is not initialized\n" );
		return;
	}

	while ( m_status != SHUTDOWN ) {
		if ( m_client.size() < static_cast<unsigned int>( MAX_NUMBER_OF_PLAYERS )) {
			// wait for an incoming client connection
			TCPsocket socket = 0;
			while( m_status != SHUTDOWN &&
			       ! ( socket = SDLNet_TCP_Accept( m_listenSocket ))) {
				SDL_Delay( 500 );
				LOG( 5 ) INFO( "ServerComm::listenForClients: listening for clients\n" );
			}
			
			if ( m_status == SHUTDOWN ) break;
			
			// retrieve client address
			IPaddress* clientAddress = SDLNet_TCP_GetPeerAddress( socket );
			if ( !clientAddress ) {
				LOG( 2 ) INFO( "ServerComm::listenForClients: SDLNet_TCP_GetPeerAddress: %s\n",
							   SDLNet_GetError() );
			}
			
			// receive client packet
			TCPPacket* packet = tcpRecvPacket( socket );
			
			if ( !packet ) continue;
			
			// in case we want to distinguish different request types
			switch ( packet->request ) { // FIXME: move variable declarations out of cases
				case TCPPacket::INFO: {
					LOG( 2 ) INFO( "ServerComm::listenForClients: received INFO request\n" );
					
					TCPInfoPacket infoPacket;
					
					// fill info packet with information
					const WopSettings* settings = WopSettings::getInstance();
					infoPacket.theme = settings->getTheme();
					infoPacket.numberPlayers = m_world->getNumberPlayers();
					tcpSendPacket( socket, &infoPacket );
					
					SDLNet_TCP_Close( socket );
					delete packet;
					break;
				}
				case TCPPacket::CONNECT: {
					LOG( 2 ) INFO( "ServerComm::listenForClients: received CONNECT request\n" );
					
					// set status to SYNC and wait for server and client to be ready
					m_status = SYNC;
					DBG( 2 ) INFO( "ServerComm::listenForClients: acquiring sync mutex\n" );
					SDL_LockMutex( m_server->getSyncMutex() );
					DBG( 2 ) INFO( "ServerComm::listenForClients: waiting on condition variable\n" );
					SDL_CondWait( m_server->getSyncCond(), m_server->getSyncMutex() );
					DBG( 2 ) INFO( "ServerComm::listenForClients: releasing sync mutex\n" );
					SDL_UnlockMutex( m_server->getSyncMutex() );
					
					TCPConnectPacket* connectPacket =
						static_cast<TCPConnectPacket*>( packet );
					Uint16 clientPort = connectPacket->port;
					delete packet;
					
					// Receive the player
					packet = tcpRecvPacket( socket );
					ASSERT( packet->request == TCPPacket::PLAYER,
							"ServerComm::listenForClients: protocol error: expected PLAYER packet\n" );
					TCPPlayerPacket* playerPacket =
						static_cast<TCPPlayerPacket*>( packet );
					Player* player = playerPacket->player;
					
					// set player id
					Uint8 playerID = m_world->getNextPlayerID();
					player->setPlayerID( playerID );
					
					if ( GameReplay* replay = m_server->getReplay() ) {
						replay->addPlayerEnter( *player );
					}
					
					// add the player to the world
					m_world->addPlayer( player );
					LOG( 1 ) INFO( "ClientComm:: added player with id %i\n",
					               player->getPlayerID() );
					
					// The port we have receive in the packet is the UDP port of the client's
					// game socket. Therefore we have to fill in this value into the client
					// address we already have, of course by transforming the value from host
					// to network byte order (i.e. serializing the value)
					IPaddress address = *clientAddress;
					Uint8 *bufferPointer = reinterpret_cast<Uint8*>( &address.port );
					Serialize<Uint16>::serialize( clientPort, bufferPointer );
					
					LOG( 2 ) INFO( "ServerComm::listenForClients: send world to the client\n" );
					TCPWorldPacket worldPacket;
					worldPacket.world = m_world;
					worldPacket.frame = m_server->getFrame();
					worldPacket.playerID = playerID;
					if ( tcpSendPacket( socket, &worldPacket ) == -1 ) {
						delete packet;
						continue;
					}
	
					LOG( 2 ) INFO( "ServerComm::listenForClients: send player to the clients\n" );
					playerPacket->frame = m_server->getFrame();
					// broadcast the new player data to the existing clients
					for ( Uint8 c = 0; c < m_client.size(); c++ ) {
						LOG( 3 ) INFO( "ServerComm::listenForClients: send player to client %c\n" );
						Client* client = m_client[c];
						if ( tcpSendPacket( client->socket, playerPacket ) == -1 ) {
							// Hmm... do we have to do anything here?
						}
					}
					
					delete packet;
					addClient( address, socket );
					m_status = DEFAULT;
					
					// signal the server main thread that we are ready to go on with the action
					DBG( 2 ) INFO( "ServerComm::listenForClients: acquiring sync mutex\n" );
					SDL_LockMutex( m_server->getSyncMutex() );
					DBG( 2 ) INFO( "ServerComm::listenForClients: signalling condition variable\n" );
					SDL_CondSignal( m_server->getSyncCond() );
					DBG( 2 ) INFO( "ServerComm::listenForClients: releasing sync mutex\n" );
					SDL_UnlockMutex( m_server->getSyncMutex() );
					
					break;
				}
				default: {
					INFO( "ServerComm::listenForClients(): unexpected packet type\n" );
					SDLNet_TCP_Close( socket );
					break;
				}
			}
		}
		else {
			SDL_Delay( 500 );
		}
	}
}

// A client is identified by its IP address and port. This method registers
// the given address as a client.

void ServerComm::addClient( const IPaddress& address, TCPsocket socket ) {
	LOG( 1 ) INFO( "ServerComm::addClient: Adding %i.%i.%i.%i:%i client %i of %i to client array\n",
	               DECODE_IP( address.host ), address.port, m_client.size(), MAX_NUMBER_OF_PLAYERS );
	Client* client = new Client();
	client->address = address;
	client->event.clear();
	client->lastReceivedFrame = (m_server->getFrame() - 1 + N_ECHO_BUFFERS) % N_ECHO_BUFFERS;
	client->serverLag = 0;
	client->clientLag = 0;
	client->socket = socket;
	m_client.push_back( client );
	SDLNet_TCP_AddSocket( m_tcpSet, client->socket );
}

// Removes the c-th client

void ServerComm::removeClient( const int c ) {
	std::vector<Client*>::iterator it = m_client.begin() + c;
	Client* client = *it;
	LOG( 1 ) INFO( "ServerComm::removeClient: Removing client %i (%i.%i.%i.%i:%i)\n", c,
	               DECODE_IP( client->address.host ), client->address.port );
	m_client.erase( it );
	SDLNet_TCP_DelSocket( m_tcpSet, client->socket );
	SDLNet_TCP_Close( client->socket );
	delete client;
	/* FIXME: do more reasonable things */
	/* Is the above FIXME still necessary? Tom */
}

// For each client we want to receive one event within a timestep. This
// method checks the game socket for incoming event packets.

void ServerComm::collectEvents() {
	// reset the up-to-date status for all clients
	unsigned int c;
	for( c = 0; c < m_client.size(); c++ ) {
		m_client[c]->upToDate = false;
	}

	// as long as there is activity on the game socket...
	// TODO: won't this cause the server to hang if there is too much traffic
	// on the game socket?
	while( SDLNet_CheckSockets( m_gameSet, 0 ) != 0 ) {
		// receive the next packet
		LOG( 5 ) INFO( "ServerComm::collectEvents: Processing event\n" );	
		if ( SDLNet_UDP_Recv( m_gameSocket, m_eventPacket ) != 1 ) {
			// no packet received? just continue
			continue;
		}
		// determine the port the packet has been sent from
		Uint16 packetPort;
		Uint8 *bufferPointer = reinterpret_cast<Uint8*>( &m_eventPacket->address.port );
		Serialize<Uint16>::deserialize( bufferPointer, packetPort );
		// check if the packet originated from a registered client
		for( c = 0; c < m_client.size() &&
		     ( m_eventPacket->address.host != m_client[c]->address.host ||
		       m_eventPacket->address.port != m_client[c]->address.port ); c++ );
		// if yes update the corresponding client status
		if( CHECK( c < m_client.size(),
		           "ServerComm::collectEvents: Received packet from unknown client (%i.%i.%i.%i:%i)\n",
		           DECODE_IP( m_eventPacket->address.host ), m_eventPacket->address.port ) ) {
			LOG( 5 ) INFO( "ServerComm::collectEvents: Received packet from client %i\n", c );

			// deserialize packet
			// m_client[c].event = *((Event*) m_eventPacket.data);
			bufferPointer = m_eventPacket->data;
			m_eventData.deserialize( bufferPointer );
			
			m_client[c]->event = m_eventData.event;
			DBG( 5 ) INFO ( "ServerComm::collectEvents: Action from client %d: %x\n", c, m_client[c]->event.get() );
			m_echoData.event[c] = m_client[c]->event.get();

			m_client[c]->upToDate = true;
			m_client[c]->serverLag = 0;
			
			// check if the client has missed an echo packet
			if ( m_client[c]->lastReceivedFrame == m_eventData.lastReceivedFrame &&
			     m_client[c]->lastReceivedFrame != m_server->getFrame() - 1 ) {
 				if ( ++m_client[c]->clientLag > m_MAX_CLIENT_LAG ) {
#ifdef LOSSY_UDP_COMMUNICATION
					if ( random() % 10 == 0 ) {
						LOG( 3 ) INFO( "ServerComm::broadcastEcho: Omitting echo frame %d for client %i\n",
								 (m_eventData.lastReceivedFrame+1) % N_ECHO_BUFFERS, c );
						continue;
					}
#endif
					// perhaps the client did not receive frame lastReceivedFrame+1
					// -> resend it
					// determine the size of the echo packet
					// TODO: we have to do this only once... find a good place to do (Comm::Comm?)
					m_echoPacket->len =
						m_echoPacket->maxlen = EchoData::getStaticSerializeBufferSize();
				
					// serialize the EchoData into the packet
					bufferPointer = m_echoPacket->data;
					m_echoBuffer[(m_eventData.lastReceivedFrame+1) % N_ECHO_BUFFERS].serialize( bufferPointer );
				
					// send the packet to the client
					m_echoPacket->address = m_client[c]->address;
					LOG( 3 ) INFO( "ServerComm::broadcastEcho: Resending frame %d to client %i\n",
								   (m_eventData.lastReceivedFrame+1) % N_ECHO_BUFFERS, c );
					SDLNet_UDP_Send( m_gameSocket, -1, m_echoPacket );
				}
			}
			else {
				m_client[c]->lastReceivedFrame = m_eventData.lastReceivedFrame;
				m_client[c]->clientLag = 0;
			}
			m_eventData.dumpToFile( "/tmp/server.log", c );
		}
	}

	// increment the lag for clients we didn't receive a packet from
	for( c = 0; c < m_client.size(); c++ ) {
		if( ! m_client[c]->upToDate ) {
			m_client[c]->serverLag++;
			m_echoData.event[c] = Event::N_A;
			LOG( 2 ) INFO( "ServerComm::collectEvents: Didn't receive packet from client %i (lag: %i)\n",
			               c, m_client[c]->serverLag );
			// if a client's lag is too large it will be forced to quit
			// TODO: perhaps there is a better way to go
			if( ! CHECK( m_client[c]->serverLag < m_MAX_SERVER_LAG,
			    "ServerComm::collectEvents: Lag of client %i too large. Sending quit signal.\n", c) ) {
				// m_echoData.event[c].setAction( Event::QUIT );
				TCPQuitPacket quitPacket;
				quitPacket.player = c;
				quitPacket.frame = m_server->getFrame();
				for ( Uint8 cl = 0; cl < m_client.size(); cl++ ) {
					if ( cl != c ) {
						LOG( 2 ) INFO( "ServerComm::collectEvents: informing client %d\n", cl );
						Client* client2 = m_client[cl];
						if ( tcpSendPacket( client2->socket, &quitPacket ) == -1 ) {
							// Hmm... do we have to do anything here?
						}
					}
				}
				
				// store quit event in replay
				// if ( GameReplay* replay = m_server->getReplay() ) {
				// 	replay->addPlayerQuit( c );
				// }
				
				// remove client, adapt loop index and continue loop
				removeClient( c );
				// m_world->removePlayer( c, m_server->getFrame() );
				// remember to remove player at the end of the frame
				m_playersToRemove.push_back( c );
				c--;
				continue;
			}
		}
	}
}

// After the event from all client have been collected they have to be
// distributed back to all clients.

void ServerComm::broadcastEcho() {
	m_echoData.frame = m_server->getFrame();
	m_echoData.status = m_status;
	m_echoData.numberClients = m_client.size();
	m_echoData.randomCounter = globalRnd.getCounter();
	
	// save the echo data in the buffer
	m_echoBuffer[m_echoData.frame] = m_echoData;
	
	// determine the size of the echo packet
	// TODO: we have to do this only once... find a good place to do (Comm::Comm?)
	m_echoPacket->len =
		m_echoPacket->maxlen = EchoData::getStaticSerializeBufferSize();

	// serialize the EchoData into the packet
	Uint8* bufferPointer = m_echoPacket->data;
	m_echoData.serialize( bufferPointer );

	// send the packet to all clients
	for( unsigned int c = 0; c < m_client.size(); c++ ) {
#ifdef LOSSY_UDP_COMMUNICATION
		if ( random() % 10 == 0 ) {
			LOG( 3 ) INFO( "ServerComm::broadcastEcho: Omitting echo frame %d for client %i\n",
			               m_echoData.frame, c );
			continue;
		}
#endif

		m_echoPacket->address = m_client[c]->address;
		LOG( 5 ) INFO( "ServerComm::broadcastEcho: Sending events to client %i (%i.%i.%i.%i:%i)\n",
		               c, DECODE_IP( m_echoPacket->address.host ), m_echoPacket->address.port );
		SDLNet_UDP_Send( m_gameSocket, -1, m_echoPacket );
	}
}

void ServerComm::handleTCPMessages() {
	// remove players
	while ( m_playersToRemove.size() > 0 ) {
		Uint8 player = m_playersToRemove.front();
		m_playersToRemove.pop_front();
	
		if ( GameReplay* replay = m_server->getReplay() ) {
			replay->addPlayerQuit( player );
		}
		
		// get player
		Player* playerPtr = m_world->getPlayer( player );
		// remove player
		m_world->removePlayer( player, m_server->getFrame() );
		// delete player
		delete playerPtr;
	}
	
	// check if there are any messages
	if ( SDLNet_CheckSockets( m_tcpSet, 0 ) > 0 ) {
		for ( unsigned int c = 0; c < m_client.size(); c++ ) {
			Client* client = m_client[c];
			if ( SDLNet_SocketReady( client->socket ) ) {
				TCPPacket* packet = tcpRecvPacket( client->socket );
				
				// A NULL packet indicates that a client has disconnected or an
				// error has occurred
				if ( !packet ) {
					LOG( 2 ) INFO( "ServerComm::handleTCPMessages: informing clients of leaving player\n" );
					TCPQuitPacket quitPacket;
					quitPacket.player = c;
					quitPacket.frame = m_server->getFrame();
					for ( Uint8 cl = 0; cl < m_client.size(); cl++ ) {
						if ( cl != c ) {
							LOG( 2 ) INFO( "ServerComm::handleTCPMessages: informing client %d\n", cl );
							Client* client2 = m_client[cl];
							if ( tcpSendPacket( client2->socket, &quitPacket ) == -1 ) {
								// Hmm... do we have to do anything here?
							}
						}
					}
					
					if ( GameReplay* replay = m_server->getReplay() ) {
						replay->addPlayerQuit( c );
					}
					
					// remove player, adapt loop index and continue loop
					removeClient( c );
					Player* player = m_world->getPlayer( c );
					m_world->removePlayer( c, m_server->getFrame() );
					delete player;
					c--;
					continue;
				}
				
				// so far only CHAT packets are acceptable
				ASSERT( packet->request == TCPPacket::CHAT,
				        "ServerComm::handleTCPMessages: protocol error: expected CHAT packet\n" );
				TCPChatPacket* chatPacket =
					static_cast<TCPChatPacket*>( packet );

				// distribute chat packet to all other clients
				for ( unsigned int r = 0; r < m_client.size(); r++ ) {
					if ( r != c ) {
						LOG( 3 ) INFO( "ServerComm::handleTCPMessages: sending chat packet to client %d\n", r );
						Client* receiver = m_client[r];
						tcpSendPacket( receiver->socket, chatPacket );
					}
				}
				
				if ( GameReplay* replay = m_server->getReplay() ) {
					replay->addChatMessage( chatPacket->message );
				}
				
				delete packet;
			}
		}
	}
}

// The following are just wrapper functions for the listen methods needed
// by the SDL thread functions

int listenForClientsWrapper( void* data ) {
	ServerComm* commPtr = static_cast<ServerComm*>( data );
	commPtr->listenForClients();
	return 0;
}

int listenForBaitPacketsWrapper( void* data ) {
	ServerComm* commPtr = static_cast<ServerComm*>( data );
	commPtr->listenForBaitPackets();
	return 0;
}
