/* $Id: weapon.hpp,v 1.6 2004/10/05 11:05:38 chfreund Exp $ */

#ifndef _WEAPON_HPP_
#define _WEAPON_HPP_

#include "serializable.hpp"
#include "serialize.hpp"
#include "audio.hpp"

class World;
class Avatar;

/*!
This class provides the basic functionality for a vanilla weapon. There are
four pure virtual methods which have to be overloaded:
- Weapon::getMaximumAmmo
- Weapon::getChargeTime
- Weapon::getReloadTime
- Weapon::reallyShoot

These methods do not have to take care of the weapon counters like remaining
ammo, reloading, ... This is done in other base classes.

To create a new weapon it is best to start from an already existing weapon
like WeaponShotgun or WeaponGrenade and alter its behavior. There are two
requirements for all new Weapon class contructors:

- \a rechargeNow must be called
- \a m_shotSample must be initialized as is the following example:
  \code
	m_shotSample = Audio::getInstance()->loadSound( "sound/weapons/shotgun/shotgun.wav" );
	\endcode
*/

class Weapon : public Serializable {
	protected:
		//! The remaining number of frames until the empty weapon is recharged.
		Uint32 m_chargeTimer;
		//! The remaining number of frames until the weapon can fire the next shot.
		Uint32 m_reloadTimer;
		//! The remaining amount of ammo.
		Uint32 m_ammo;
		//! This sample is played after a successful shot.
		Mix_Chunk* m_shotSample;

	public:
		//! \return The maximum amount of ammo.
		virtual Uint32 getMaximumAmmo() const = 0;

		//! \return The required number of frames for recharging an completely empty weapon.
		virtual Uint32 getChargeTime() const = 0;

		//! \return The required number of frames for preparing the next shot.
		virtual Uint32 getReloadTime() const = 0;

		//! This method should check whether a shot is possible. If shooting is
		//! possible, place particles and/or other weapon objects.
		//! \note Checking or decreasing the ammo, charge or reload counters is
		//! not necessary.
		//! \return Return \c true if the weapon could be fired successfully.
		//! Otherwise \c false.
		virtual bool reallyShoot( World* world, Avatar* avatar ) const = 0;

		//! All basic things for handling the weapon counters are handled here.
		//! If shooting is possible according to these counters, Weapon::reallyShoot
		//! will be called and the shot sample will be played.
		virtual void shoot( World* world, Avatar* avatar );

		//! Takes care of the weapon counters.
		virtual void doTimestep() {
			if ( m_chargeTimer > 0 && --m_chargeTimer == 0 ) {
				m_ammo = getMaximumAmmo();
			}

			if ( m_reloadTimer > 0 ) {
				m_reloadTimer--;
			}
		}

		//! \return \c true if weapon is recharging; \c false otherwise.
		virtual bool isCharging() const { return m_chargeTimer > 0; }

		//! \return The current amount of ammo.
		virtual Uint32 getCurrentAmmo() { return m_ammo; }

		//! \return The current charging state between 0 (not charged) and 100
		//! (completely charged).
		virtual Uint32 getChargingPercentage() {;
			return 100 - ( 100 * m_chargeTimer ) / getChargeTime();
		}

		//! Sets all weapon counters to a completely charged weapon.
		virtual void rechargeNow() {
			m_ammo = getMaximumAmmo();
			m_reloadTimer = m_chargeTimer = 0;
		}

		virtual Uint32 getSerializeBufferSize() const {
			return 3 * Serialize<Uint32>::sizeOf();
		}

		virtual void serialize( Uint8*& bufferPointer ) const {
			Serialize<Uint32>::serialize( m_ammo, bufferPointer );
			Serialize<Uint32>::serialize( m_reloadTimer, bufferPointer );
			Serialize<Uint32>::serialize( m_chargeTimer, bufferPointer );
		}

		virtual void deserialize( Uint8*& bufferPointer ) {
			Serialize<Uint32>::deserialize( bufferPointer, m_ammo );
			Serialize<Uint32>::deserialize( bufferPointer, m_reloadTimer );
			Serialize<Uint32>::deserialize( bufferPointer, m_chargeTimer );
		}
		
};

#endif // _WEAPON_HPP_
