/* $Id: world.cpp,v 1.138 2005/04/29 15:03:04 uwe Exp $ */

#include "world.hpp"

#include <new>
#include <algorithm>

#include "wopsettings.hpp"
#include "theme.hpp"
#include "event.hpp"
#include "udpdata.hpp"
#include "random.hpp"
#include "avatarworm.hpp"
#include "explosion.hpp"
#include "blueglow.hpp"
#include "goal.hpp"
#include "shieldcorona.hpp"
#include "invisiblecorona.hpp"
#include "smoke.hpp"
#include "grenade.hpp"
#include "mine.hpp"
#include "missile.hpp"
#include "homingmissile.hpp"
#include "flag.hpp"
#include "rope.hpp"
#include "hook.hpp"
#include "ball.hpp"
#include "bonushealth.hpp"
#include "bonusshield.hpp"
#include "bonusinvisible.hpp"
#include "bonusfuel.hpp"
#include "bonushomingvirus.hpp"
#include "spritesequence.cpp"
#include "spriteset.cpp"
#include "wopsprites.cpp"

/**********************************************************/

World::World( int sizeX, int sizeY )
	: m_particles( *this ),
	  m_bonusmanager( this, globalRnd.getUint32() ),
	  m_spriteInterface( NULL ),
	  m_nextPlayerID( 0 )
{
	Object::m_worldPointer = this;
	Scorekeeper::getInstance()->initialize( this );
	m_map = new ServerMap( sizeX, sizeY );
	m_mutex = SDL_CreateMutex();
}

/**********************************************************/

World::~World() {
	LOG( 1 ) INFO( "World::~World: starting destructor\n" );

	reset();
	
	delete m_map;
	
	SDL_DestroyMutex( m_mutex );
	
	LOG( 1 ) INFO( "World::~World: done\n" );
}

/**********************************************************/

bool World::createFromTheme()
{
	SDL_mutexP( m_mutex );
	
	const WopSettings* settings = WopSettings::getInstance();
	
	//////////////////////////////////////////////
	// load the theme
	//////////////////////////////////////////////

	String themePath( settings->getData() );
	themePath += '/';
	themePath += THEME_CONFIG_PATH;
	Theme theme( settings->getTheme(), globalRnd.getUint32() & 0xffff );
	ASSERT( theme.load((char*)themePath, THEME_CONFIG),
	        "World::createFromTheme: could not load theme \"%s/"
	        THEME_CONFIG ":%s\"\n",
	        (char*)themePath, (char*)theme.getName() );

	//////////////////////////////////////////////
	// build the map from the theme
	//////////////////////////////////////////////

	// if the loaded theme defines its own map size, set it now
	if( theme.definesOwnSize() ) {
		// adapt the global settings
		if( WopSettings::getInstance()->getWidth() != theme.getWidth() ) {
			WopSettings::getInstance()->setWidth( theme.getWidth() ); }
		if( WopSettings::getInstance()->getHeight() != theme.getHeight() ) {
			WopSettings::getInstance()->setHeight( theme.getHeight() ); }
		// build new map, if necessary
		if( m_map->getSizeX() != (Uint32)theme.getWidth() ||
		    m_map->getSizeY() != (Uint32)theme.getHeight()   ) {
			LOG(2) INFO( "World::createFromTheme: theme \"%s\" defines "
			             "its own map size %dx%d.\n      Overwriting "
			             "current map size %dx%d.\n",
			             (char*)theme.getName(),
			             theme.getWidth(), theme.getHeight(),
			             m_map->getSizeX(), m_map->getSizeY() );
			// delete old map and create new one
			delete m_map;
			ASSERT( NULL != (m_map = new ServerMap(theme.getWidth(),
			                                       theme.getHeight())),
			        "World::createFromTheme: could not create server map "
			        "of size %dx%d\n", theme.getWidth(), theme.getHeight() );
		}
	}
	// create the map from the theme
	ServerMap* serverMap = dynamic_cast<ServerMap*>( m_map );
	ASSERT( serverMap, "World::createFromTheme: no ServerMap to create theme\n" );
	// first load eventual configuration for initial objects
	if( theme.hasObjectConfigPath() ) {
		String objectPath( settings->getData() );
		objectPath += '/';
		objectPath += theme.getObjectConfigPath();
	}
	// build map from theme
	serverMap->createFromTheme( theme );

	//////////////////////////////////////////////
	// load the settings for the bonus object handling
	//////////////////////////////////////////////

	String bonusPath( settings->getData() );
	bonusPath += '/';
	bonusPath += theme.getBonusConfigPath();
	ASSERT( m_bonusmanager.loadSettings(bonusPath),
	       "World::createFromTheme: could not load bonus configuration "
	       "from file \"%s\"\n", (char*)bonusPath );
	bonusPath.reset();

	SDL_mutexV( m_mutex );
	
	return true;
}

/**********************************************************/

void World::setupGameMode() {
	Scorekeeper::getInstance()->setupGameMode();
}

/**********************************************************/

void World::optimizeForClient( SDL_PixelFormat* format, Uint32 flags ) {
	if ( !dynamic_cast<ClientMap*>( m_map ) ) {
		ServerMap* serverMap = dynamic_cast<ServerMap*>( m_map );
		ASSERT( serverMap, "World::optimizeForClient: map type unknown\n" );
		LOG( 4 ) INFO( "World::optimizeForClient: Creating optimized map for the client\n" );
		ClientMap* optimizedMap = new ClientMap( *serverMap, format, flags );
		ASSERT( optimizedMap, "World::optimizeForClient: could not optimize map\n" );
		LOG( 4 ) INFO( "World::optimizeForClient: Deleting original server map\n" );
		delete m_map;
		m_map = optimizedMap;
	}
	else {
		INFO( "World::optimizeForClient: map is already optimized\n" );
	}
}

/**********************************************************/

void World::addPlayer( Player* p )
{
	SDL_mutexP( m_mutex );
	
	// create a new avatar object, attach it to the player object
	p->attachAvatar( dynamic_cast<Avatar*>(newObject(p->getAvatarType())) );

	// initialize new avatar
	p->getAvatar()->spawnAvatar();

	// update the sprite selection
	// IMPORTANT:
	//   this cannot be done in the constructor of class Avatar
	//   since updateSpriteSelection() will cause a call of
	//   the pure virtual function getID(), which is implemented
	//   only in the derived classes.
	p->getAvatar()->updateSpriteSelection();
	// add the player to the world's player list
	m_player.push_back( p );

	printPlayerVectorInfo();
	Scorekeeper::getInstance()->resetPlayerScore( p->getPlayerID() );

	SDL_mutexV( m_mutex );
}

/**********************************************************/

void World::removePlayer( const Uint8 c, const Uint8 frame )
{
	SDL_mutexP( m_mutex );
	
	LOG( 2 ) INFO( "World::removePlayer: removing player %d in frame %d\n", c, frame );
	// remove the avatar from the world's list of 
	// collidable objects so that it is not drawn any more
	//Object *avatar = m_collidables.swap_unhook( m_player[c]->getAvatar() );
	Object* avatar = m_player[c]->getAvatar();

	const vector<CollidableObject*>::iterator avatarIt =
	              std::find( m_collidables.begin(), m_collidables.end(), avatar );
	DBG( 1 ) ASSERT( avatarIt != m_collidables.end(), "World::removePlayer: player could not be found\n" );
	m_collidables.erase( avatarIt );

	// check, if the colored instance can be deleted
	bool InstanceIsFree = true;
	for( Uint8 p = 0; p < m_player.size(); p++ ) {
		if( p != c &&
		       m_player[p]->getAvatar()->getID()
		       == avatar->getID()
		    && m_player[p]->getAvatar()->getColorInstance()
		       == avatar->getColorInstance() )
		{
			InstanceIsFree = false;
			break;
		}
	}
	
	// remove the avatar's colored instance from the sprite object
	if( InstanceIsFree ) {
		((WopSpritesWorldInterface*)getSpriteInterface())->
			removeColoredInstance( *avatar );
	}
	
	// detach the avatar from the player, who leaves the world
	// NOTE: the order is important here (1) detach (2) delete
	m_player[c]->detachAvatar();
	delete avatar;

	// remove the player from the world's player list
	m_player.erase( m_player.begin() + c );

	printPlayerVectorInfo();

	SDL_mutexV( m_mutex );
}

/**********************************************************/

Player* World::getPlayerByID( const Uint8 id ) const {
	SDL_mutexP( m_mutex );
	
	for ( Uint8 p = 0; p < m_player.size(); p++ ) {
		Player* player = m_player[p];
		if ( player->getPlayerID() == id ) {
			SDL_mutexV( m_mutex );
			return player;
		}
	}
	
	SDL_mutexV( m_mutex );
	return 0;
}

/**********************************************************/

Object* World::newObject( const int objectID )
{
	SDL_mutexP( m_mutex );
	
	Object* object = NULL;

	// create object
	switch( objectID ){
		case AVATAR_WORM  :
			object = NEW AvatarWorm();
			break;
		case GRENADE :
			object = NEW Grenade();
			break;
		case MINE :
			object = NEW Mine();
			break;
		case BALL :
			object = NEW Ball();
			break;
		case MISSILE :
			object = NEW Missile();
			break;
		case HOMING_MISSILE :
			object = NEW HomingMissile();
			break;
		case BONUS_HEALTH :
			object = NEW BonusHealth();
			break;
		case BONUS_SHIELD :
			object = NEW BonusShield();
			break;
		case BONUS_INVISIBLE :
			object = NEW BonusInvisible();
			break;
		case BONUS_FUEL :
			object = NEW BonusFuel();
			break;
		case BONUS_HOMING_VIRUS :
			object = NEW BonusHomingVirus();
			break;
		case EXPLOSION_15 :
			object = NEW Explosion( EXPLOSION_15 - SMALLEST_EXPLOSION );
			break;
		case EXPLOSION_40 :
			object = NEW Explosion( EXPLOSION_40 - SMALLEST_EXPLOSION );
			break;
		case EXPLOSION_60 :
			object = NEW Explosion( EXPLOSION_60 - SMALLEST_EXPLOSION );
			break;
		case EXPLOSION_80 :
			object = NEW Explosion( EXPLOSION_80 - SMALLEST_EXPLOSION );
			break;
		case SMOKE_15 :
			object = NEW Smoke( SMOKE_15 - SMALLEST_SMOKE );
			break;
		case SMOKE_30 :
			object = NEW Smoke( SMOKE_30 - SMALLEST_SMOKE );
			break;
		case BLUEGLOW_25 :
			object = NEW BlueGlow( BLUEGLOW_25 - SMALLEST_BLUEGLOW );
			break;
		case GOAL_50 :
			object = NEW Goal();
			break;
		case SHIELD_CORONA :
			object = NEW ShieldCorona();
			break;
		case INVISIBLE_CORONA :
			object = NEW InvisibleCorona();
			break;
		case ROPE :
			object = NEW Rope();
			break;
		case HOOK :
			object = NEW Hook();
			break;
		default: break;
	}

	DBG( 1 ) ASSERT( object != NULL, "World::newObject: non-valid "
	                 "object ID %d\n", objectID );

	// update the sequence length. IMPORTANT:
	//   this cannot be done in the constructor of class Object
	//   since updateSpriteSelection() will cause a call of
	//   the pure virtual function getID(), which is implemented
	//   only in the derived classes.
	object->updateSequenceLength();
	// insert it in the correct vector
	if( Object::isCollidable( objectID ) ) {
		m_collidables.push_back( dynamic_cast<CollidableObject*>( object ));
	} else {
		m_noncollidables.push_back( dynamic_cast<NonCollidableObject*>( object ));
	}
	
	SDL_mutexV( m_mutex );
	
	return object;
}

/**********************************************************/

void World::deleteObject( Object* const object ) {
	unhookObject( object);
	delete object;
}

/**********************************************************/

void World::unhookObject( const Object* const object ) {
	SDL_mutexP( m_mutex );

	if( object->isCollidable() ) {
		const vector<CollidableObject*>::iterator collIt =
		            std::find( m_collidables.begin(), m_collidables.end(), object );
		if( collIt != m_collidables.end() ) m_collidables.erase( collIt );
		else CHECK( false, "World::deleteObject: collidable object could not be found\n" );
	}
	else {
		const vector<NonCollidableObject*>::iterator noncollIt =
		      std::find( m_noncollidables.begin(), m_noncollidables.end(), object );
		if( noncollIt != m_noncollidables.end() ) m_noncollidables.erase( noncollIt );
		else  CHECK( false, "World::deleteObject: noncollidable object could not be found\n" );
	}
	
	SDL_mutexV( m_mutex );
}

/**********************************************************/

void World::applyEvents( const EchoData* echoData ) {
	SDL_mutexP( m_mutex );
	
	// Process the actions of all players
	Uint8 numberPlayers = m_player.size();

	for( Uint8 p = 0; p < numberPlayers; p++ ) {
		const Event& event = echoData->event[p];
		
		Player* player = getPlayer(p);
		Avatar* avatar = player->getAvatar();

		if( ! avatar->isLiving() ) {
			// check if player wants to respawn
			if ( event.testAnyActions( Event::SPAWN ) ) {
				avatar->spawnAvatar();
			}
			
			continue;
		}

		// do we have an up-to-date event packet?
		if( ! event.testAnyActions( Event::N_A ) ) {

			// reset moving, jumping, and digging
			avatar->setState( Avatar::MOVING,   false );
			avatar->setState( Avatar::JUMPING,  false );
			avatar->setState( Avatar::DIGGING,  false );

			// handle aiming
			if( ! event.testAnyActions( Event::UP | Event::DOWN )) {
				avatar->setAimingDirection( 0 );
			}
			else {
				if( event.testAnyActions( Event::DOWN )) {
					avatar->setAimingDirection( 1 );
				}
				else { // this must be Event::UP
					avatar->setAimingDirection( -1 );
				}
			}

			// handle moving
			if( event.testAnyActions( Event::RIGHT )) {
				avatar->setState( Avatar::MOVING, true );
				avatar->setDirection( Avatar::RIGHT );
			}

			if( event.testAnyActions( Event::LEFT )) {
				avatar->setState( Avatar::MOVING, true );
				avatar->setDirection( Avatar::LEFT );
			}

			// handle jumping
			if( event.testAnyActions( Event::JUMP ))
				avatar->setState( Avatar::JUMPING, true );

			// handle digging
			if( event.testAnyActions( Event::DIG ) )
				avatar->setState( Avatar::DIGGING, true );

			// handle weapon selection
			const Uint32 weapon = event.getWeapon();
			if ( weapon >= 1 && weapon <= Player::m_MAX_WEAPONS-Player::m_IGNORE_N_LAST_WEAPONS /* because of rope weapons */)
				player->setActiveWeapon( weapon-1 );
		}

		avatar->setState( Avatar::ROPE_ON,  false  );
		avatar->setState( Avatar::ROPE_OFF, false  );
		avatar->setState( Avatar::ROPE_IN,  false  );
		avatar->setState( Avatar::ROPE_OUT, false  );
		avatar->setState( Avatar::ROPE_REL, false  );

		// handle rope events
		if( event.testAnyActions( Event::ROPE_ON ) ) {
			avatar->setState( Avatar::ROPE_ON, true );
		}
		if( event.testAnyActions( Event::ROPE_OFF ) ) {
			avatar->setState( Avatar::ROPE_OFF, true );
		}
		if( event.testAnyActions( Event::ROPE_IN ) ) {
			avatar->setState( Avatar::ROPE_IN, true );
		}
		if( event.testAnyActions( Event::ROPE_OUT ) ) {
			avatar->setState( Avatar::ROPE_OUT, true );
		}
		if( event.testAnyActions( Event::ROPE_REL ) ) {
			avatar->setState( Avatar::ROPE_REL, true );
		}

		avatar->setState( Avatar::SHOOTING, false );
		if( event.testAnyActions( Event::SHOOT ) )
				avatar->setState( Avatar::SHOOTING, true );
		
	}

	SDL_mutexV( m_mutex );
}

/**********************************************************/

void World::damageObject( const Particles::ParticleData& p ) {
	SDL_mutexP( m_mutex );
	
	CollidableObject* hitObject = getCollidableAt( ROUND( p.pos.x ),
	                                               ROUND( p.pos.y ) );

	if( hitObject ) {
		const Player* const shooter = getPlayerByID( p.owner );
		if( shooter ) {
			LOG( 5 ) INFO( "World::damageObject: player '%s' hit something\n",
			               static_cast<char*>( shooter->getName() ));
			hitObject->applyDamage( p );
		}
		else {
			LOG( 1 ) CHECK( false, "World::damageObject: cannot determine shooter\n" );
		}

	}

	SDL_mutexV( m_mutex );
}

/**********************************************************/

void World::doTimestep()
{
	SDL_mutexP( m_mutex );
	
	// update particles
	m_particles.doTimestep();

	// update collidable objects
	for( unsigned int i = 0; i < m_collidables.size(); i++ ) {
		m_collidables[i]->update();
	}

	// update noncollidable objects
	for( unsigned int i = 0; i < m_noncollidables.size(); i++ ) {
		m_noncollidables[i]->update();
	}

	// update the bonus creation
	m_bonusmanager.updatePlacement();
	
	// update the players' weapons
	for ( Uint8 p = 0; p < m_player.size(); p++ ) {
		Player* player = m_player[p];
		for ( int i = 0; i < Player::m_MAX_WEAPONS; i++ ) {
			player->getWeapon( i )->doTimestep();
		}
	}

	// check scoring
	//m_scorekeeper.checkGoals();

	SDL_mutexV( m_mutex );
}

/**********************************************************/

void World::draw( SDL_Surface* surface, ClientWopSprites& sprites,
                  SDL_Rect& mapRect ) const {
	SDL_mutexP( m_mutex );
	
	// draw the noncollidable objects and ropes, part 1
	for( unsigned int i = 0; i < m_noncollidables.size(); i++ ) {
		NonCollidableObject* obj = m_noncollidables[i];
		// ropes must be drawn seperately, since they do not use sprites
		if( obj->getID() == ROPE ) {
			if ( SDL_MUSTLOCK( surface ) )
				SDL_LockSurface( surface );
			Rope* const rope = dynamic_cast<Rope*>( obj );
			// draw the rope line
			Graphics::drawLine( surface,
			                    rope->getStartIntPositionX() - mapRect.x,
			                    rope->getStartIntPositionY() - mapRect.y,
			                    rope->getEndIntPositionX()   - mapRect.x,
			                    rope->getEndIntPositionY()   - mapRect.y,
			                    rope->getColor() );
			// for charged ropes we must draw a lightning
			if( rope->isCharged() ) {
				Vector sparkPosOld( rope->getStartPosition() ),
						 sparkPosNew,
						 ropeVector( rope->getEndPosition() - sparkPosOld );

				for( real pos = 0.1 + 0.1 * localRnd.getNormedReal(); pos < 1.0;
					  pos += 0.1 + 0.1 * localRnd.getNormedReal() ) {
					sparkPosNew = rope->getStartPosition() + ropeVector * pos +
									  localRnd.getVector( 30.0 * (0.25 - (0.5-pos)*(0.5-pos)) );
					// draw this lightning line
					Graphics::drawLine( surface,
					                    ROUND( sparkPosOld.x ) - mapRect.x,
					                    ROUND( sparkPosOld.y ) - mapRect.y,
					                    ROUND( sparkPosNew.x ) - mapRect.x,
					                    ROUND( sparkPosNew.y ) - mapRect.y,
					                    localRnd.getColor( 0x9999ff, 0xeeeeff ));
					sparkPosOld = sparkPosNew;
				}
				// draw finish lightning line
				Graphics::drawLine( surface,
				                    ROUND( sparkPosOld.x ) - mapRect.x,
				                    ROUND( sparkPosOld.y ) - mapRect.y,
				                    rope->getEndIntPositionX()   - mapRect.x,
				                    rope->getEndIntPositionY()   - mapRect.y,
				                    localRnd.getColor( 0x9999ff, 0xeeeeff ));
				// delete the charge flag (it will be set
				// anew, if we hold on electrifying something)
				rope->setCharged( false );
			}
			if ( SDL_MUSTLOCK( surface ) )
				SDL_UnlockSurface( surface );
		// the ordinary objects are drawn using sprites
		} else {
			if( obj->isVisible() && obj->getDrawBeforeCollidableObjects() )
				sprites.draw( *obj,
				              obj->getIntPositionX() - mapRect.x,
				              obj->getIntPositionY() - mapRect.y,
				              surface );
		}
	}
	
	// draw the collidable objects
	for( unsigned int i = 0; i < m_collidables.size(); i++ ) {
		CollidableObject* obj = m_collidables[i];
		if( obj->isVisible() )
			sprites.draw( *obj,
			              obj->getIntPositionX() - mapRect.x,
			              obj->getIntPositionY() - mapRect.y,
			              surface );
	}

	// draw the particles
	m_particles.draw( surface, mapRect );
	
	// draw the noncollidable objects, part 2
	for( unsigned int i = 0; i < m_noncollidables.size(); i++ ) {
		NonCollidableObject* obj = m_noncollidables[i];
		if( obj->isVisible() && ! obj->getDrawBeforeCollidableObjects() )
			sprites.draw( *obj,
			              obj->getIntPositionX() - mapRect.x,
			              obj->getIntPositionY() - mapRect.y,
			              surface );
	}
	
	
	SDL_mutexV( m_mutex );
}

/**********************************************************/
/*
World* World::createAndDeserialize( Uint8*& bufferPointer ) {
	ASSERT( false, "World::createAndDeserialize: don't use me, I'm broken\n" );

	World* newObject = NEW World( 1, 1 );
	
	newObject->getMap()->deserialize( bufferPointer );

	Uint8 numberPlayers;
	Serialize<Uint8>::deserialize( bufferPointer, numberPlayers );
	for ( Uint8 p = 0; p < numberPlayers; p++ ) {
		Player* player = NEW Player();
		player->deserialize( bufferPointer );
		newObject->m_player.push_back( player );
	}

	return newObject;
}
*/
/**********************************************************/

// The common methods related to serialization. For a description look
// at the comments in map.hpp

Uint32 World::getSerializeBufferSize() const
{
	Uint32 bufferSize = 0;

	ADD_TAG_SIZE( bufferSize );

	// server settings
	bufferSize += WopSettings::getInstance()->getSerializeBufferSize();
	
	ADD_TAG_SIZE( bufferSize );

	// map
	ServerMap* serverMap = dynamic_cast<ServerMap*>( m_map );
	ASSERT( serverMap, "World::getSerializeBufferSize: map is already optimized\n" );
	bufferSize += serverMap->getSerializeBufferSize();

	ADD_TAG_SIZE( bufferSize );

	// collidable objects
	bufferSize += Serialize<Uint32>::sizeOf(); // vector length
	for( Uint32 i = 0; i < m_collidables.size(); i++ ) {
		ADD_TAG_SIZE( bufferSize );
		bufferSize += Serialize<Uint32>::sizeOf();  // object ID
		bufferSize += m_collidables[i]->getSerializeBufferSize();
	}
	
	ADD_TAG_SIZE( bufferSize );

	// noncollidable objects
	bufferSize += Serialize<Uint32>::sizeOf(); // vector length
	for( Uint32 i = 0; i < m_noncollidables.size(); i++ ) {
		ADD_TAG_SIZE( bufferSize );
		bufferSize += Serialize<Uint32>::sizeOf();  // object ID
		bufferSize += m_noncollidables[i]->getSerializeBufferSize();
	}
	
	ADD_TAG_SIZE( bufferSize );

	// player list
	bufferSize += Serialize<Uint8>::sizeOf();
	for ( Uint8 p = 0; p < m_player.size(); p++ ) {
		bufferSize += Serialize<Uint32>::sizeOf(); // index of the avatar
		bufferSize += m_player[p]->getSerializeBufferSize();
	}
	
	ADD_TAG_SIZE( bufferSize );

	// particles
	bufferSize += m_particles.getSerializeBufferSize();

	ADD_TAG_SIZE( bufferSize );

	// random
	bufferSize += globalRnd.getSerializeBufferSize();

	ADD_TAG_SIZE( bufferSize );

	// scorekeeper
	bufferSize += Scorekeeper::getInstance()->getSerializeBufferSize();

	ADD_TAG_SIZE( bufferSize );

	// bonus manager
	bufferSize += m_bonusmanager.getSerializeBufferSize();
	
	ADD_TAG_SIZE( bufferSize );

	return bufferSize;
}

/**********************************************************/

void World::serialize( Uint8*& bufferPointer ) const
{
	// this macro must stay in the first line
	START_SERIALIZED_SIZE_CHECK( bufferPointer );
	
	SERIALIZE_TAG( bufferPointer )

	// server settings
	WopSettings::getInstance()->serializeServerSettings( bufferPointer );

	SERIALIZE_TAG( bufferPointer )

	// map
	ServerMap* serverMap = dynamic_cast<ServerMap*>( m_map );
	ASSERT( serverMap, "World::getSerializeBufferSize: map is already optimized\n" );
	serverMap->serialize( bufferPointer );

	SERIALIZE_TAG( bufferPointer )
	
	// bonus manager
	// Must be DEserialized before all collidable objects, because
	// the bonus manager also serializes the bonus limits and the
	// deserialization of collidable objects might already create
	// bonus objects, which register at the bonus manager.
	m_bonusmanager.serialize( bufferPointer );

	SERIALIZE_TAG( bufferPointer )
	// collidable objects
	// Must he serialized before the player list, since in the
	// deserialization process the players will be linked with their
	// avatars, that must be present at this time.
	Serialize<Uint32>::serialize( static_cast<Uint32>(m_collidables.size()),
	                              bufferPointer );
	for( Uint32 i = 0; i < m_collidables.size(); i++ ) {
		Serialize<Sint32>::serialize( // object ID
		                   static_cast<Sint32>(m_collidables[i]->getID()),
		                   bufferPointer );
		m_collidables[i]->serialize( bufferPointer ); // object itself
		SERIALIZE_TAG( bufferPointer )
	}

	SERIALIZE_TAG( bufferPointer )

	// noncollidable objects
	Serialize<Uint32>::serialize( static_cast<Uint32>(m_noncollidables.size()),
	                              bufferPointer );
	for( Uint32 i = 0; i < m_noncollidables.size(); i++ ) {
		SERIALIZE_TAG( bufferPointer )
		Serialize<Sint32>::serialize( // object ID
		                   static_cast<Sint32>(m_noncollidables[i]->getID()),
		                   bufferPointer );
		m_noncollidables[i]->serialize( bufferPointer ); // object itself
	}

	SERIALIZE_TAG( bufferPointer )

	// player list
	Serialize<Uint8>::serialize( static_cast<Uint8>( m_player.size() ),
	                             bufferPointer );
	for ( Uint8 p = 0; p < m_player.size(); p++ ) {
		// find index of the attached avatar
		for( Uint32 i = 0; i < m_collidables.size(); i++ ) {
			if( m_collidables[i] == m_player[p]->getAvatar() ) {
				Serialize<Uint32>::serialize( i, bufferPointer );
				break;
			}
			DBG( 1 ) ASSERT( i < m_collidables.size()-1,
			                 "World::serialize: could not find avatar for "
			                 "player \"%s\"\n",
			                 m_player[p]->getName().getString() );
		}
		m_player[p]->serialize( bufferPointer );
	}

	SERIALIZE_TAG( bufferPointer )

	// particles
	m_particles.serialize( bufferPointer );

	SERIALIZE_TAG( bufferPointer )

	// random
	globalRnd.serialize( bufferPointer );

	SERIALIZE_TAG( bufferPointer )

	// scorekeeper
	Scorekeeper::getInstance()->serialize( bufferPointer );

	SERIALIZE_TAG( bufferPointer )
	
	// this macro must stay in the last line
	END_SERIALIZED_SIZE_CHECK( bufferPointer, World );
}

/**********************************************************/

void World::deserialize( Uint8*& bufferPointer )
{
	DESERIALIZE_TAG( bufferPointer )

	// server settings
	WopSettings::getInstance()->deserializeServerSettings( bufferPointer );
	Scorekeeper::getInstance()->setGameMode( WopSettings::getInstance()->getGameMode() );
	DESERIALIZE_TAG( bufferPointer )

	// deserialize the map
	DBG( 2 ) INFO( "World::deserialize: deserializing of 'map' started\n" );
	ServerMap* serverMap = dynamic_cast<ServerMap*>( m_map );
	ASSERT( serverMap, "World::getSerializeBufferSize: map is already optimized\n" );
	serverMap->deserialize( bufferPointer );
	DBG( 2 ) INFO( "World::deserialize: deserializing of 'map' ended\n" );
	DESERIALIZE_TAG( bufferPointer )

	// bonus manager
	// Must be DEserialized before all collidable objects, because
	// the bonus manager also serializes the bonus limits and the
	// deserialization of collidable objects might already create
	// bonus objects, which register at the bonus manager.
	DBG( 2 ) INFO( "World::deserialize: deserializing of 'bonus manager' started\n" );
	m_bonusmanager.deserialize( bufferPointer );
	DBG( 2 ) INFO( "World::deserialize: deserializing of 'bonus manager' ended\n" );
	DESERIALIZE_TAG( bufferPointer )
	
	// collidable objects
	// Must be deserialized befor the player list, because the player
	// list will read an index for each player, which is the index of
	// each player's avatar in the object list.
	DBG( 2 ) INFO( "World::deserialize: deserializing of 'collidable objects' started\n" );
	Uint32 numObjects = 0;
	Sint32 objectID   = -1;
	Serialize<Uint32>::deserialize( bufferPointer, numObjects );
	for( Uint32 i = 0; i < numObjects; i++ ) {
		DBG( 2 ) INFO( "World::deserialize: deserializing 'collidable object' %i of %i\n", i, numObjects );
		Serialize<Sint32>::deserialize( bufferPointer, objectID );
		// get object from the world
		Object* object = newObject( objectID );
		// deserialize object
		object->deserialize( bufferPointer );
		DESERIALIZE_TAG( bufferPointer )
	}

	DBG( 2 ) INFO( "World::deserialize: deserializing of 'collidable objects' ended\n" );
	DESERIALIZE_TAG( bufferPointer )

	// noncollidable objects
	DBG( 2 ) INFO( "World::deserialize: deserializing of 'noncollidable objects' started\n" );
	numObjects = 0;
	Serialize<Uint32>::deserialize( bufferPointer, numObjects );
	for( Uint32 i = 0; i < numObjects; i++ ) {
		DBG( 2 ) INFO( "World::deserialize: deserializing of 'noncollidable objects': object %i of %i\n", i, numObjects );
		DESERIALIZE_TAG( bufferPointer )
		Serialize<Sint32>::deserialize( bufferPointer, objectID );
		// get object from the world
		Object* object = newObject( objectID );
		// deserialize object
		object->deserialize( bufferPointer );
	}

	AttachableRegistrar::getInstance()->startPointerConversion();
	AttachableRegistrar::getInstance()->reset();

	DBG( 2 ) INFO( "World::deserialize: deserializing of 'noncollidable objects' ended\n" );
	DESERIALIZE_TAG( bufferPointer )

	// deserialize the player list
	DBG( 2 ) INFO( "World::deserialize: deserializing of 'players' started\n" );
	Uint8 numberPlayers;
	Serialize<Uint8>::deserialize( bufferPointer, numberPlayers );
	m_player.clear();
	for ( Uint8 p = 0; p < numberPlayers; p++ ) {
		// read index of the player's avatar in noncollidable object list
		Uint32 avatarIndex;
		Serialize<Uint32>::deserialize( bufferPointer, avatarIndex );
		// create player
		Player* player = NEW Player();
		player->deserialize( bufferPointer );
		m_player.push_back( player );
		// link player with its avatar
		player->attachAvatar( static_cast<Avatar*>(m_collidables[avatarIndex]) );
		DBG( 1 ) ASSERT( m_collidables[avatarIndex]->getID() <= LAST_AVATAR,
		                 "World::deserialize: linked player with a non-avatar "
		                 "object (ID %d == \"%s\")\n",
		                 m_collidables[avatarIndex]->getID(),
		                 ObjectIDString[m_collidables[avatarIndex]->getID()] );
	}

	DBG( 2 ) INFO( "World::deserialize: deserializing of 'players' ended\n" );
	DESERIALIZE_TAG( bufferPointer )

	// particles
	DBG( 2 ) INFO( "World::deserialize: deserializing of 'particles' started\n" );
	m_particles.deserialize( bufferPointer );
	DBG( 2 ) INFO( "World::deserialize: deserializing of 'particles' ended\n" );
	DESERIALIZE_TAG( bufferPointer )

	// random
	DBG( 2 ) INFO( "World::deserialize: deserializing of 'random' started\n" );
	globalRnd.deserialize( bufferPointer );
	DBG( 2 ) INFO( "World::deserialize: deserializing of 'random' ended\n" );
	DESERIALIZE_TAG( bufferPointer )

	// scorekeeper
	DBG( 2 ) INFO( "World::deserialize: deserializing of 'scorekeeper' started\n" );
	Scorekeeper::getInstance()->deserialize( bufferPointer );
	DBG( 2 ) INFO( "World::deserialize: deserializing of 'scorekeeper' ended\n" );
	DESERIALIZE_TAG( bufferPointer )
}

void World::reset() {
	SDL_mutexP( m_mutex );
	
	// detach all avatars, they will die after that anyway
	for( Uint8 p = 0; p < m_player.size(); p++ ) {
		m_player[p]->detachAvatar();
		delete m_player[p];
	}

	// delete all collidable objects
	for( unsigned int i = 0; i < m_collidables.size(); i++ )
		delete m_collidables[i];
	m_collidables.clear();
	
	// delete all noncollidable objects
	for( unsigned int i = 0; i < m_noncollidables.size(); i++ )
		delete m_noncollidables[i];
	m_noncollidables.clear();
	
	m_particles.reset();

	Scorekeeper::getInstance()->reset();

	m_bonusmanager.reset();
	
	// Note: The map m_map is resetted automatically during the next
	// deserialization

	SDL_mutexV( m_mutex );
}

/**********************************************************/

void World::dump() {
	SDL_mutexP( m_mutex );
	
	INFO( "World status:\n" );
	for ( int i = 0; i < int( m_player.size() ); i++ ) {
		Player* p = m_player[i];
		Avatar* a = p->getAvatar();
		INFO( "\tPlayer %d position: (%f, %f)\n", i, a->getPositionX(), a->getPositionY() );
	}
	
	SDL_mutexV( m_mutex );
}

/**********************************************************/

void World::printPlayerVectorInfo() {
	INFO( "World::printPlayerVectorInfo:\n" );
	INFO( "\tnPlayers: %i\n", m_player.size() );
	for ( int i = 0; i < int( m_player.size() ); i++ ) {
		Player* p = m_player[i];
		INFO( "\tplayer %d: id: %i; name: %s\n", i, p->getPlayerID(),
		      p->getName().getString() );
	}
}

/**********************************************************/

CollidableObject* World::getCollidableAt( const int x, const int y ) {
	SDL_mutexP( m_mutex );
	
	DBG( 4 ) {
		for( unsigned int o = 0; o < m_collidables.size(); o++ ) {
			if( m_collidables[o]->testHit( x, y ) ) {
				SDL_mutexV( m_mutex );
				return m_collidables[o];
			}
		}
	}
	else {
		CollidableObject* foundObject = NULL;
		for( unsigned int o = 0; o < m_collidables.size(); o++ ) {
			if( m_collidables[o]->testHit( x, y ) ) {
				if( foundObject == NULL ) {
					foundObject = m_collidables[o];
				}
				else {
					CHECK( false, "World::getCollidableAt: found more "
					       "than one object at (%i, %i): %s (%i, %i, %i, %i) and "
					       "%s (%i, %i, %i, %i)\n", x, y,
					       ObjectIDString[foundObject->getID()],
					       foundObject->getIntPositionX(), foundObject->getIntPositionY(),
					       foundObject->getCollRectWidth(), foundObject->getCollRectHeight(),
					       ObjectIDString[m_collidables[o]->getID()],
					       m_collidables[o]->getIntPositionX(), m_collidables[o]->getIntPositionY(),
					       m_collidables[o]->getCollRectWidth(), m_collidables[o]->getCollRectHeight() );
					String comment;
					comment.format( "World::getCollidableAt: found more than one object at (%i, %i)", x, y );
					m_map->saveFlags( comment );
				}
			}
		}
		if( foundObject ) {
			SDL_mutexV( m_mutex );
			return foundObject;
		}
	}

	LOG( 1 ) CHECK( false, "World::getCollidableAt: cannot find any collidable "
	                       "object at (%i, %i)\n", x, y );
	
	SDL_mutexV( m_mutex );
	
	return NULL;
}

/**********************************************************/
